from __future__ import annotations

from typing import TYPE_CHECKING
from typing import final

from yatbaf.types import Message

from .abc import TelegramMethod

if TYPE_CHECKING:
    from yatbaf.enums import ParseMode
    from yatbaf.types import InlineKeyboardMarkup
    from yatbaf.types import LinkPreviewOptions
    from yatbaf.types import MessageEntity
    from yatbaf.typing import NoneInt


@final
class EditMessageText(TelegramMethod[Message | bool]):
    """See :meth:`yatbaf.bot.Bot.edit_message_text`"""

    text: str
    chat_id: str | int | None = None
    message_id: NoneInt = None
    inline_message_id: NoneInt = None
    parse_mode: ParseMode | None = None
    entities: list[MessageEntity] | None = None
    link_preview_options: LinkPreviewOptions | None = None
    reply_markup: InlineKeyboardMarkup | None = None



from yatbaf.router import OnMessage


async def _guard_func(_):  # noqa: U101
    return False


def test_init_guard():
    router = OnMessage(guards=[_guard_func])
    assert _guard_func in router._guards


def test_add_guard():
    router = OnMessage()
    router.add_guard(_guard_func)
    assert len(router._guards) == 1
    assert _guard_func in router._guards


def test_add_guard_duplicate():
    router = OnMessage()
    func = object()

    router.add_guard(func)
    router.add_guard(func)
    assert len(router._guards) == 1
    assert func in router._guards


def test_guard_decorator(router):

    @router.guard
    async def func(_):  # noqa: U101
        pass

    assert len(router._guards) == 1
    assert func in router._guards



from __future__ import annotations

__all__ = (
    "OnMessage",
    "OnEditedMessage",
    "OnChannelPost",
    "OnEditedChannelPost",
    "OnMessageReaction",
    "OnMessageReactionCount",
    "OnInlineQuery",
    "OnChosenInlineResult",
    "OnCallbackQuery",
    "OnShippingQuery",
    "OnPreCheckoutQuery",
    "OnPoll",
    "OnPollAnswer",
    "OnMyChatMember",
    "OnChatMemeber",
    "OnChatJoinRequest",
    "OnChatBoost",
    "OnRemovedChatBoost",
)

import logging
from itertools import count
from typing import TYPE_CHECKING
from typing import Final
from typing import Literal
from typing import TypeAlias
from typing import TypeVar
from typing import cast
from typing import final
from typing import overload

from .abc import AbstractRouter
from .exceptions import FrozenInstanceError
from .handler import Handler
from .middleware import Middleware
from .types import CallbackQuery
from .types import ChatBoostRemoved
from .types import ChatBoostUpdated
from .types import ChatJoinRequest
from .types import ChatMemberUpdated
from .types import ChosenInlineResult
from .types import InlineQuery
from .types import Message
from .types import MessageReactionCountUpdated
from .types import MessageReactionUpdated
from .types import Poll
from .types import PollAnswer
from .types import PreCheckoutQuery
from .types import ShippingQuery
from .typing import UpdateT
from .utils import ensure_unique
from .warnings import warn_duplicate

if TYPE_CHECKING:
    from collections.abc import Sequence

    from .models import UpdateInfo
    from .typing import Filter
    from .typing import GuardCallable
    from .typing import HandlerCallable
    from .typing import HandlerMiddleware
    from .typing import MiddlewareType
    from .typing import RouterCallable
    from .typing import RouterMiddleware
    M = TypeVar("M", bound=RouterMiddleware | HandlerMiddleware)
else:
    M = TypeVar("M")

log = logging.getLogger(__name__)
_router_count = count(1).__next__

T = TypeVar("T")
UnionMiddlewareType: TypeAlias = (
    "Middleware[HandlerCallable[T]] "
    "| Middleware[RouterCallable[T]] "
)
InitMiddlewareType: TypeAlias = (
    "UnionMiddlewareType[T] "
    "| HandlerMiddleware[T]"
)


class BaseRouter(AbstractRouter[UpdateT]):
    """Common behaviour for :class:`Router` and
    :class:`~yatbaf.dispatcher.Dispatcher`.
    """

    __slots__ = (
        "_guards",
        "_parent",
        "_middleware",
    )

    _middleware_stack: RouterCallable[UpdateT]

    def __init__(
        self,
        guards: Sequence[GuardCallable[UpdateT]] | None = None,
        middleware: Sequence[InitMiddlewareType[UpdateT]] | None = None,
        routers: Sequence[AbstractRouter[UpdateT]] | None = None,
        handlers: Sequence[Handler[UpdateT]] | None = None,
    ) -> None:
        """
        :param guards: *Optional.* A sequence of :class:`~yatbaf.typing.Guard`.
        :param middleware: *Optional.* A sequence of :class:`~yatbaf.middleware.Middleware`
            or :class:`~yatbaf.typing.HandlerMiddleware`.
        :param routers: *Optional.* A sequence of :class:`~yatbaf.abc.AbstractRouter`.
        :param handlers: *Optional.* A sequence of :class:`~yatbaf.typing.Handler`.
        """  # noqa: E501
        self._parent = None
        self._guards: list[GuardCallable[UpdateT]] = ensure_unique(guards or [])
        self._middleware = ensure_unique([  # yapf: disable  # type: ignore[assignment]  # noqa: E501
            Middleware(m, is_handler=True)  # type: ignore[misc]
            if not isinstance(m, Middleware) else m
            for m in (middleware or [])
        ])
        self._register_routers(routers or [])
        self._register_handlers(handlers or [])

    def _register_middleware(
        self, middleware: Sequence[UnionMiddlewareType[UpdateT]]
    ) -> None:
        outer = []
        inner = []
        for obj in middleware:
            if obj.is_handler:
                inner.append(obj)
            else:
                outer.append(obj)
        for cls in reversed(cast("list[RouterMiddleware[UpdateT]]", outer)):
            self._middleware_stack = cls(self._middleware_stack)
        self._middleware = cast(
            "list[Middleware[HandlerCallable[UpdateT]]]", inner
        )

    async def _check_guards(self, update: UpdateT) -> bool:
        for func in self._guards:
            if not await func(update):
                return False
        return True


class Router(BaseRouter[UpdateT]):
    """Base class for routers."""

    __slots__ = (
        "_routers",
        "_sort_filters",
        "_skip_with_nested",
        "_update_type",
        "_handlers",
        "_name",
        "_frozen",
        "_middleware_stack",
    )

    def __init__(
        self,
        *,
        handlers: Sequence[Handler[UpdateT]] | None = None,
        middleware: Sequence[InitMiddlewareType[UpdateT]] | None = None,
        routers: Sequence[AbstractRouter[UpdateT]] | None = None,
        guards: Sequence[GuardCallable[UpdateT]] | None = None,
        name: str | None = None,
        sort_filters: bool = True,
        skip_with_nested: bool = False,
    ) -> None:
        """
        :param handlers: *Optional.* A sequence of :class:`~yatbaf.typing.Handler`.
        :param middleware: *Optional.* A sequence of :class:`~yatbaf.middleware.Middleware`
            or :class:`~yatbaf.typing.HandlerMiddleware`.
        :param routers: *Optional.* A sequence of :class:`~yatbaf.abc.AbstractRouter`.
        :param guards: *Optional.* A sequence of :class:`~yatbaf.typing.Guard`.
        :param name: *Optional.* Router name.
        :param sort_filters: *Optional.* Sort handler filters by priority.
        :param skip_with_nested: *Optional.* Guard will also skip nested routers.
        """  # noqa: E501
        self._sort_filters: Final[bool] = sort_filters
        self._skip_with_nested: Final[bool] = skip_with_nested
        self._routers: list[AbstractRouter[UpdateT]] = []
        self._handlers: list[Handler[UpdateT]] = []
        self._name: Final[str] = name if name else f"router-{_router_count()}"
        self._frozen = False
        self._middleware_stack = self.__resolve

        super().__init__(
            routers=routers,
            handlers=handlers,
            guards=guards,
            middleware=middleware,
        )

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}[name={self._name}]>"

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Router) and (  # yapf: disable
            other is self or (
                other._update_type == self._update_type
                and other._handlers == self._handlers
                and other._guards == self._guards
                and other._middleware == self._middleware
                and other._routers == self._routers
            )
        )

    @property
    def name(self) -> str:
        """Router name."""
        return self._name

    def _register_routers(
        self, routers: Sequence[AbstractRouter[UpdateT]]
    ) -> None:
        for router in routers:
            self.add_router(router)

    def _register_handlers(self, handlers: Sequence[Handler[UpdateT]]) -> None:
        for handler in handlers:
            self.add_handler(handler)

    def add_guard(self, func: GuardCallable[UpdateT], /) -> None:
        """Add a new guard function.

        :param func: :class:`Guard <yatbaf.typing.Guard>` function.
        :raises FrozenInstanceError: If you try to register a Guard after Bot
            object has been initialized.
        """
        if self._frozen:
            raise FrozenInstanceError(
                "It is not possible to add a new Guard at runtime "
                "after Bot object has been initialized."
            )

        if func not in self._guards:
            self._guards.append(func)

    def add_middleware(
        self,
        func: HandlerMiddleware[UpdateT] | RouterMiddleware[UpdateT],
        *,
        is_handler: bool = True,
        is_local: bool = False,
    ) -> None:
        """Add a new middleware.

        Add a new middleware which will be applied to all handlers in router.
        Function must return a new Handler functoin. Usage::

            def middleware(handler: Handler) -> Handler:
                async def wrapper(update: UpdateT) -> None:
                    await handler(update)
                return wrapper

            router.add_middleware(middleware)

        :param func: :class:`Middleware <yatbaf.types.Middleware>` function.
        :param is_handler: *Optional.* Pass ``False`` if it is an router
            middleware.
        :param is_local: *Optional.* Pass ``True`` if it is a middleware only
            for handlers in current router. For handler middleware only
            (``is_handler``=``True``).
        :raises FrozenInstanceError: If you try to register a Middleware after
            Bot object has been initialized.
        """
        if self._frozen:
            raise FrozenInstanceError(
                "It is not possible to add a new Middleware at runtime "
                "after Bot object has been initialized."
            )

        middleware = cast(
            "Middleware[RouterCallable[UpdateT]] | Middleware[HandlerCallable[UpdateT]]",  # noqa: E501
            Middleware(
                func,
                is_handler=is_handler,
                is_local=is_local,
            )
        )

        # skip duplicate
        if middleware not in self._middleware:
            self._middleware.append(middleware)  # type: ignore[arg-type]

    @overload
    def add_handler(self, handler: Handler[UpdateT]) -> None:
        ...

    @overload
    def add_handler(
        self,
        handler: HandlerCallable[UpdateT],
        *,
        filters: Sequence[Filter[UpdateT]] | None = None,
        middleware: Sequence[HandlerMiddleware[UpdateT]] | None = None,
        sort_filters: bool | None = None,
        any_filter: bool = False,
    ) -> None:
        ...

    def add_handler(
        self,
        handler: HandlerCallable[UpdateT] | Handler[UpdateT],
        *,
        filters: Sequence[Filter[UpdateT]] | None = None,
        middleware: Sequence[HandlerMiddleware[UpdateT]] | None = None,
        sort_filters: bool | None = None,
        any_filter: bool = False,
    ) -> None:
        """Use this method to register a new handler.

        :param handler: :class:`~yatbaf.handler.Handler` instance or function.
        :param filters: *Optional.* A sequence of :class:`~yatbaf.typing.Filter`.
        :param middleware: *Optional.* A sequence of :class:`~yatbaf.typing.HandlerMiddleware`.
        :param sort_filters: *Optional.* Pass ``False`` if you want to use your
            filter order. Default to :attr:`Router.sort_filters`.
        :param any_filter: Pass ``True`` if matching one of the filters is enough.
        :raises FrozenInstanceError: If you try to register a Handler after Bot
            object has been initialized.
        """  # noqa: E501
        if self._frozen:
            raise FrozenInstanceError(
                f"{self!r} is frozen. It is not possible to add a new Handler "
                "at runtime after Bot object has been initialized."
            )

        if not isinstance(handler, Handler):
            handler = Handler(
                fn=handler,
                update_type=self._update_type,
                filters=filters,
                middleware=middleware,
                sort_filters=(  # yapf: disable
                    self._sort_filters
                    if sort_filters is None
                    else sort_filters
                ),
                any_filter=any_filter,
            )

        if handler.update_type != self._update_type:
            raise ValueError(
                f"Wrong handler type! Cannot add {handler!r} to {self!r}"
            )

        parent = handler._parent
        if parent is not None and parent is not self:
            raise ValueError(f"{handler!r} alredy registered in {parent!r}")

        # skip duplicate
        if parent or handler in self._handlers:
            warn_duplicate(handler, self)
            return

        self._handlers.append(handler)
        handler._parent = self

    @overload
    def __call__(  # yapf: disable
        self, __fn: HandlerCallable[UpdateT], /
    ) -> HandlerCallable[UpdateT]:
        ...

    @overload
    def __call__(
        self,
        *,
        filters: Sequence[Filter[UpdateT]] | None = None,
        middleware: Sequence[HandlerMiddleware[UpdateT]] | None = None,
        sort_filters: bool | None = None,
        any_filter: bool = False,
    ) -> HandlerMiddleware[UpdateT]:
        ...

    def __call__(
        self,
        __fn: HandlerCallable[UpdateT] | None = None,
        *,
        filters: Sequence[Filter[UpdateT]] | None = None,
        middleware: Sequence[HandlerMiddleware[UpdateT]] | None = None,
        sort_filters: bool | None = None,
        any_filter: bool = False,
    ) -> HandlerMiddleware[UpdateT] | HandlerCallable[UpdateT]:
        """Handler decorator.

        See :meth:`add_handler`.

        Use this decorator to register a new handler::

            @router
            async def handler(update):
                ...

            @router(filters=[filter1, filter2])
            async def handler(update):
                ...
        """

        def wrapper(fn: HandlerCallable[UpdateT]) -> HandlerCallable[UpdateT]:
            self.add_handler(
                handler=fn,
                filters=filters,
                middleware=middleware,
                sort_filters=sort_filters,
                any_filter=any_filter,
            )
            return fn

        if __fn is not None:
            return wrapper(__fn)
        return wrapper

    def add_router(self, router: AbstractRouter[UpdateT], /) -> None:
        """Use this method to add a nested router.

        .. warning::

            The nested router must handle the same type of updates as the
            current one.

        :param router: :class:`~yatbaf.abc.AbstractRouter` instance.
        :raises ValueError: If ``router`` already registered in another router
            or router type is different.
        :raises FrozenInstanceError: If you try to register a Router after Bot
            object has been initialized.
        """
        if self._frozen:
            raise FrozenInstanceError(
                f"{self!r} is frozen. It is not possible to add a new Router "
                "at runtime after Bot object has been initialized."
            )

        if self._update_type != router._update_type:
            raise ValueError(
                f"Wrong router type! Cannot add {router!r} to {self!r}"
            )

        if router is self:
            raise ValueError(f"It is not possible to add {router!r} to itself.")

        parent = router._parent
        if parent is not None and parent is not self:
            raise ValueError(f"{router!r} already registered in {parent!r}")

        if parent or router in self._routers:
            warn_duplicate(router, self)
            return

        self._routers.append(router)
        router._parent = self

    def _find_handler(self, update: UpdateT) -> Handler[UpdateT] | None:
        result: Handler[UpdateT] | None = None
        for handler in self._handlers:
            if handler._match(update):
                if not handler._is_fallback():
                    return handler
                result = handler
        return result

    async def __resolve(self, update: UpdateInfo[UpdateT], /) -> bool:
        content = update.content
        if await self._check_guards(content):
            if (handler := self._find_handler(content)) is not None:
                await handler(content)
                return True

        # guard failed
        elif self._skip_with_nested:
            return False

        # try to find the handler in nested routers
        for router in self._routers:
            if await router._resolve(update):
                return True

        return False

    async def _resolve(self, update: UpdateInfo[UpdateT], /) -> bool:
        return await self._middleware_stack(update)

    @overload
    def middleware(
        self, __fn: HandlerMiddleware[UpdateT]
    ) -> HandlerMiddleware[UpdateT]:
        ...

    @overload
    def middleware(
        self,
        *,
        is_handler: bool = False,
    ) -> MiddlewareType[RouterMiddleware[UpdateT]]:
        ...

    @overload
    def middleware(
        self,
        *,
        is_local: bool,
        is_handler: bool = True,
    ) -> MiddlewareType[HandlerMiddleware[UpdateT]]:  # noqa: E501
        ...

    def middleware(
        self,
        __fn: HandlerMiddleware[UpdateT] | None = None,
        *,
        is_handler: bool = True,
        is_local: bool = False,
    ) -> (  # yapf: disable
        MiddlewareType[HandlerMiddleware[UpdateT]]
        | MiddlewareType[RouterMiddleware[UpdateT]]
        | HandlerMiddleware[UpdateT]
    ):
        """Middleware decorator.

        Use this decorator to register a middleware for router::

            @router.middleware
            def middleware(handler):
                async def wrapper(update):
                    await handler(update)
                return wrapper
        """

        def wrapper(
            __fn: RouterMiddleware[UpdateT] | HandlerMiddleware[UpdateT]
        ) -> RouterMiddleware[UpdateT] | HandlerMiddleware[UpdateT]:
            self.add_middleware(
                __fn,
                is_handler=is_handler,
                is_local=is_local,
            )
            return __fn

        if __fn is not None:
            return cast("HandlerMiddleware[UpdateT]", wrapper(__fn))
        return cast(
            "MiddlewareType[RouterMiddleware[UpdateT]] | MiddlewareType[HandlerMiddleware[UpdateT]]",  # noqa: E501
            wrapper
        )

    def guard(self, fn: GuardCallable[UpdateT], /) -> GuardCallable[UpdateT]:
        """Guard decorator.

        Use this decorator to register a guard for router::

            @router.guard
            async def guard(update):
                return True
        """
        self.add_guard(fn)
        return fn

    def _on_registration(self) -> None:
        self._frozen = True
        self._register_middleware(self._middleware)

        middleware = []
        local_middleware = []
        for obj in self._middleware:
            if obj.is_local:
                local_middleware.append(obj)
            else:
                middleware.append(obj)
        self._middleware = middleware + local_middleware

        for handler in self._handlers:
            handler._on_registration()
        self._middleware = middleware

        for router in self._routers:
            router._on_registration()


@final
class OnMessage(Router[Message]):
    """message router.

    See :attr:`Update.message <yatbaf.types.update.Update.message>`
    """

    __slots__ = ()
    _update_type: Literal["message"] = "message"


@final
class OnEditedMessage(Router[Message]):
    """edited_message router.

    See :attr:`Update.edited_message <yatbaf.types.update.Update.edited_message>`
    """  # noqa: E501

    __slots__ = ()
    _update_type: Literal["edited_message"] = "edited_message"


@final
class OnChannelPost(Router[Message]):
    """channel_post router.

    See :attr:`Update.channel_post <yatbaf.types.update.Update.channel_post>`
    """

    __slots__ = ()
    _update_type: Literal["channel_post"] = "channel_post"


@final
class OnEditedChannelPost(Router[Message]):
    """edited_channel_post router.

    See :attr:`Update.edited_channel_post <yatbaf.types.update.Update.edited_channel_post>`
    """  # noqa: E501

    __slots__ = ()
    _update_type: Literal["edited_channel_post"] = "edited_channel_post"


@final
class OnMessageReaction(Router[MessageReactionUpdated]):
    """message_reaction router.

    See :attr:`Update.message_reaction <yatbaf.types.update.Update.message_reaction>`
    """  # noqa: E501

    __slots__ = ()
    _update_type: Literal["message_reaction"] = "message_reaction"


@final
class OnMessageReactionCount(Router[MessageReactionCountUpdated]):
    """message_reaction router.

    See :attr:`Update.message_reaction_count <yatbaf.types.update.Update.message_reaction_count>`
    """  # noqa: E501

    __slots__ = ()
    _update_type: Literal["message_reaction_count"] = "message_reaction_count"


@final
class OnInlineQuery(Router[InlineQuery]):
    """inline_query router.

    See :attr:`Update.inline_query <yatbaf.types.update.Update.inline_query>`
    """

    __slots__ = ()
    _update_type: Literal["inline_query"] = "inline_query"


@final
class OnChosenInlineResult(Router[ChosenInlineResult]):
    """chosen_inline_result router.

    See :attr:`Update.chosen_inline_result <yatbaf.types.update.Update.chosen_inline_result>`
    """  # noqa: E501

    __slots__ = ()
    _update_type: Literal["chosen_inline_result"] = "chosen_inline_result"


@final
class OnCallbackQuery(Router[CallbackQuery]):
    """callback_query router.

    See :attr:`Update.callback_query <yatbaf.types.update.Update.callback_query>`
    """  # noqa: E501

    __slots__ = ()
    _update_type: Literal["callback_query"] = "callback_query"


@final
class OnShippingQuery(Router[ShippingQuery]):
    """shipping_query router.

    See :attr:`Update.shipping_query <yatbaf.types.update.Update.shipping_query>`
    """  # noqa: E501

    __slots__ = ()
    _update_type: Literal["shipping_query"] = "shipping_query"


@final
class OnPreCheckoutQuery(Router[PreCheckoutQuery]):
    """pre_checkout_query router.

    See :attr:`Update.pre_checkout_query <yatbaf.types.update.Update.pre_checkout_query>`
    """  # noqa: E501

    __slots__ = ()
    _update_type: Literal["pre_checkout_query"] = "pre_checkout_query"


@final
class OnPoll(Router[Poll]):
    """poll router.

    See :attr:`Update.poll <yatbaf.types.update.Update.poll>`
    """

    __slots__ = ()
    _update_type: Literal["poll"] = "poll"


@final
class OnPollAnswer(Router[PollAnswer]):
    """poll_answer router.

    See :attr:`Update.poll_answer <yatbaf.types.update.Update.poll_answer>`
    """

    __slots__ = ()
    _update_type: Literal["poll_answer"] = "poll_answer"


@final
class OnMyChatMember(Router[ChatMemberUpdated]):
    """my_chat_member router.

    See :attr:`Update.my_chat_member <yatbaf.types.update.Update.my_chat_member>`
    """  # noqa: E501

    __slots__ = ()
    _update_type: Literal["my_chat_member"] = "my_chat_member"


@final
class OnChatMemeber(Router[ChatMemberUpdated]):
    """chat_member router.

    See :attr:`Update.chat_member <yatbaf.types.update.Update.chat_member>`
    """

    __slots__ = ()
    _update_type: Literal["chat_member"] = "chat_member"


@final
class OnChatJoinRequest(Router[ChatJoinRequest]):
    """chat_join_request router.

    See :attr:`Update.chat_join_request <yatbaf.types.update.Update.chat_join_request>`
    """  # noqa: E501

    __slots__ = ()
    _update_type: Literal["chat_join_request"] = "chat_join_request"


@final
class OnChatBoost(Router[ChatBoostUpdated]):
    """chat_boost router.

    See :attr:`Update.chat_boost <yatbaf.types.update.Update.chat_boost>`
    """  # noqa: E501

    __slots__ = ()
    _update_type: Literal["chat_boost"] = "chat_boost"


@final
class OnRemovedChatBoost(Router[ChatBoostRemoved]):
    """removed_chat_boost router.

    See :attr:`Update.removed_chat_boost <yatbaf.types.update.Update.removed_chat_boost>`
    """  # noqa: E501

    __slots__ = ()
    _update_type: Literal["removed_chat_boost"] = "removed_chat_boost"


router_map: dict[str, type[BaseRouter]] = {
    "message": OnMessage,
    "edited_message": OnEditedMessage,
    "channel_post": OnChannelPost,
    "edited_channel_post": OnEditedChannelPost,
    "message_reaction": OnMessageReaction,
    "message_reaction_count": OnMessageReactionCount,
    "inline_query": OnInlineQuery,
    "chosen_inline_result": OnChosenInlineResult,
    "callback_query": OnCallbackQuery,
    "shipping_query": OnShippingQuery,
    "pre_checkout_query": OnPreCheckoutQuery,
    "poll": OnPoll,
    "poll_answer": OnPollAnswer,
    "my_chat_member": OnMyChatMember,
    "chat_member": OnChatMemeber,
    "chat_join_request": OnChatJoinRequest,
    "chat_boost": OnChatBoost,
    "removed_chat_boost": OnRemovedChatBoost,
}



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import final

from .abc import TelegramMethod

if TYPE_CHECKING:
    from yatbaf.types import ReactionType
    from yatbaf.typing import NoneBool


@final
class SetMessageReaction(TelegramMethod[bool]):
    """See: :meth:`~yatbaf.bot.Bot.set_message_reaction`."""

    chat_id: int | str
    message_id: int
    reaction: list[ReactionType] | None = None
    is_big: NoneBool = None



from __future__ import annotations

__all__ = (
    "ResponseError",
    "ResponseOk",
    "Request",
    "HTTPResponse",
    "UpdateInfo",
)

from dataclasses import dataclass
from typing import Any
from typing import Generic
from typing import TypeVar
from typing import final

from msgspec import Struct

from .types import ResponseParameters
from .typing import NoneStr
from .typing import UpdateT

T = TypeVar("T")


@final
class Request(Struct):
    """Request object"""

    url: str
    """URL."""

    method: str
    """Telegram method name."""

    data: dict[str, Any] | None = None
    """Request POST data."""

    files: dict[str, Any] | None = None
    """Request files."""

    content: bytes | None = None
    """Json (bytes) content"""

    headers: dict[str, str] | None = None
    """Request headers."""


@final
class HTTPResponse(Struct, Generic[T]):
    """This object represents HTTP response."""

    status_code: int
    """Request status code."""

    content: T
    """Response content."""


class ApiResponse(Struct):
    """Base object for api response.

    See: https://core.telegram.org/bots/api#making-requests
    """

    ok: bool
    """Request status."""


@final
class ResponseOk(ApiResponse, Generic[T]):
    """This object represents successful api response."""

    result: T
    """Result of the query."""

    description: NoneStr = None
    """Human-readable description of the result."""


@final
class ResponseError(ApiResponse):
    """This object represents unsuccessful api response."""

    error_code: int
    """HTTP status code."""

    description: str
    """Human-readable description of the error."""

    parameters: ResponseParameters | None = None
    """See :class:`yatbaf.types.ResponseParameters`."""


@dataclass(frozen=True)
class UpdateInfo(Generic[UpdateT]):
    id: int
    """Update id"""
    name: str
    """Update type"""
    content: UpdateT
    """Update content"""



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import ClassVar
from typing import final

from yatbaf.types import Message

from .abc import TelegramMethodWithMedia

if TYPE_CHECKING:
    from yatbaf.types import InlineKeyboardMarkup
    from yatbaf.types import InputMedia
    from yatbaf.typing import NoneInt


@final
class EditMessageMedia(TelegramMethodWithMedia[Message | bool]):
    """See :meth:`yatbaf.bot.Bot.edit_message_media`"""

    media: InputMedia
    chat_id: str | int | None = None
    message_id: NoneInt = None
    inline_message_id: NoneInt = None
    reply_markup: InlineKeyboardMarkup | None = None

    __meth_media_fields__: ClassVar[tuple[str, ...]] = ("media",)



from __future__ import annotations

from typing import final

from yatbaf.typing import NoneInt
from yatbaf.typing import NoneStr

from .abc import TelegramType


@final
class Voice(TelegramType):
    """This object represents a voice note.

    See: https://core.telegram.org/bots/api#voice
    """

    file_id: str
    """Identifier for this file, which can be used to download or reuse the file."""  # noqa: E501

    file_unique_id: str
    """Unique identifier for this file, which is supposed to be the same over
    time and for different bots. Can't be used to download or reuse the file.
    """

    duration: int
    """Duration of the audio in seconds as defined by sender."""

    mime_type: NoneStr = None
    """*Optional.* MIME type of the file as defined by sender"""

    file_size: NoneInt = None
    """*Optional.* File size in bytes."""



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import final

from .abc import TelegramMethod

if TYPE_CHECKING:
    from yatbaf.typing import NoneStr


@final
class SetCustomEmojiStickerSetThumbnail(TelegramMethod[bool]):
    """See :meth:`yatbaf.bot.Bot.set_custom_emoji_sticker_set_thumbnail`"""

    name: str
    custom_emoji_id: NoneStr = None



from typing import final

from .abc import TelegramType


@final
class GeneralForumTopicHidden(TelegramType):
    """This object represents a service message about General forum topic
    hidden in the chat. Currently holds no information.

    See: https://core.telegram.org/bots/api#generalforumtopichidden
    """



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import final

from .abc import TelegramMethod

if TYPE_CHECKING:
    from yatbaf.typing import NoneStr


@final
class EditForumTopic(TelegramMethod[bool]):
    """See :meth:`yatbaf.bot.Bot.edit_forum_topic`"""

    chat_id: str | int
    message_thread_id: int
    name: NoneStr = None
    icon_custom_emoji_id: NoneStr = None



from __future__ import annotations

from typing import final

from msgspec import field

from yatbaf.typing import NoneStr

from .abc import TelegramType
from .location import Location
from .user import User


@final
class ChosenInlineResult(TelegramType):
    """Represents a result of an inline query that was chosen by the user and
    sent to their chat partner.

    See: https://core.telegram.org/bots/api#choseninlineresult
    """

    result_id: int
    """The unique identifier for the result that was chosen."""

    from_: User = field(name="from")
    """The user that chose the result."""

    query: str
    """The query that was used to obtain the result."""

    location: Location | None = None
    """*Optional.* Sender location, only for bots that require user location."""

    inline_message_id: NoneStr = None
    """*Optional.* Identifier of the sent inline message. Available only if
    there is an inline keyboard attached to the message. Will be also received
    in callback queries and can be used to edit the message.
    """



from __future__ import annotations

from typing import final

from .abc import TelegramMethod


@final
class SetStickerEmojiList(TelegramMethod[bool]):
    """See :meth:`yatbaf.bot.Bot.set_sticker_emoji_list`"""

    sticker: str
    emoji_list: list[str]



from __future__ import annotations

from typing import final

from yatbaf.enums import StickerType
from yatbaf.typing import NoneBool
from yatbaf.typing import NoneInt
from yatbaf.typing import NoneStr

from .abc import TelegramType
from .file import File
from .mask_position import MaskPosition
from .photo_size import PhotoSize


@final
class Sticker(TelegramType):
    """This object represents a sticker.

    See: https://core.telegram.org/bots/api#sticker
    """

    file_id: str
    """
    Identifier for this file, which can be used to download or reuse the file.
    """

    file_unique_id: str
    """Unique identifier for this file, which is supposed to be the same over
    time and for different bots. Can't be used to download or reuse the file.
    """

    type: StickerType
    """Type of the sticker."""

    width: int
    """Sticker width."""

    height: int
    """Sticker height."""

    is_animated: bool
    """True, if the sticker is animated."""

    is_video: bool
    """True, if the sticker is a video sticker."""

    thumbnail: PhotoSize | None = None
    """*Optional.* Sticker thumbnail in the .WEBP or .JPG format."""

    emoji: NoneStr = None
    """*Optional.* Emoji associated with the sticker."""

    set_name: NoneStr = None
    """*Optional.* Name of the sticker set to which the sticker belongs."""

    premium_animation: File | None = None
    """*Optional.* For premium regular stickers, premium animation for the
    sticker.
    """

    mask_position: MaskPosition | None = None
    """*Optional.* For mask stickers, the position where the mask should be
    placed.
    """

    custom_emoji_id: NoneStr = None
    """*Optional.* For custom emoji stickers, unique identifier of the custom
    emoji.
    """

    needs_repainting: NoneBool = None
    """*Optional.* True, if the sticker must be repainted to a text color in
    messages, the color of the Telegram Premium badge in emoji status, white
    color on chat photos, or another appropriate color in other places.
    """

    file_size: NoneInt = None
    """*Optional.* File size in bytes."""



from __future__ import annotations

from typing import final

from .abc import TelegramType


@final
class ForumTopicClosed(TelegramType):
    """This object represents a service message about a forum topic closed in
    the chat. Currently holds no information.

    See: https://core.telegram.org/bots/api#forumtopicclosed
    """



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import final

from yatbaf.types import ForumTopic

from .abc import TelegramMethod

if TYPE_CHECKING:
    from yatbaf.enums import IconColor
    from yatbaf.typing import NoneStr


@final
class CreateForumTopic(TelegramMethod[ForumTopic]):
    """See :meth:`yatbaf.bot.Bot.create_forum_topic`"""

    chat_id: str | int
    name: str
    icon_color: IconColor | None = None
    icon_custom_emoji_id: NoneStr = None



from __future__ import annotations

from typing import final

from yatbaf.types import ChatMember

from .abc import TelegramMethod


@final
class GetChatAdministrators(TelegramMethod[list[ChatMember]]):
    """See :meth:`yatbaf.bot.Bot.get_chat_administrators`"""

    chat_id: str | int



from __future__ import annotations

__all__ = [
    "parse_command",
    "decode_content",
    "build_decoder",
]

import re
from typing import TYPE_CHECKING
from typing import TypeVar

from msgspec import DecodeError
from msgspec.json import Decoder

from .exceptions import InvalidTokenError
from .exceptions import JSONDecodeError
from .models import ResponseError
from .models import ResponseOk
from .types import Update

if TYPE_CHECKING:
    from collections.abc import Callable
    from collections.abc import Sequence

    from .methods.abc import TelegramMethod
    from .typing import ResultModelT

    F = TypeVar("F", bound=Callable)

    def lru_cache(
        maxsize: int = 128,  # noqa: U100
        typed: bool = False,  # noqa: U100
    ) -> Callable[[F], F]:
        pass
else:
    from functools import lru_cache

T = TypeVar("T")


def parse_command(text: str) -> str | None:
    """Parse bot command.

    :param text: Message text.
    """

    if not text.startswith("/") or text == "/":
        return None

    return text.removeprefix("/").split("@")[0].split(maxsplit=1)[0].lower()


def decode_content(content: bytes, decoder: Decoder[T]) -> T:
    """Decode content.

    :meta private:
    :param content: Raw content.
    :param decoder: Decoder object.
    """

    try:
        return decoder.decode(content)
    except DecodeError as error:
        raise JSONDecodeError(
            message=str(error),
            raw_content=content,
        ) from None


@lru_cache(maxsize=10)
def build_decoder(model: type[T]) -> Decoder[T]:
    return Decoder(model)


def decode_response(
    method: TelegramMethod[ResultModelT],
    content: bytes,
    decoder: Decoder[ResponseOk[ResultModelT]] | None = None,
) -> ResponseOk[ResultModelT]:
    """Decode response content.

    :meta private:
    :param method: TelegramMethod object.
    :param content: Response raw content.
    :param decoder: *Optional.* Decoder object.
    """

    if decoder is None:
        model = method._get_result_model()
        decoder = build_decoder(ResponseOk[model])  # type: ignore[valid-type]
    return decode_content(content, decoder)


def decode_webhook(content: bytes) -> Update:
    """Decode webhook content.

    :meta private:
    :param content: Request raw content.
    """

    decoder = build_decoder(Update)
    return decode_content(content, decoder)


def decode_error(content: bytes) -> ResponseError:
    """Decode response error.

    :meta private:
    """

    decoder = build_decoder(ResponseError)
    return decode_content(content, decoder)


DEEPLINK_PAYLOAD_PATTERN = re.compile(r"^[a-zA-Z0-9_-]{,64}$")
USERNAME_PATTERN = re.compile(r"^[a-zA-Z0-9_]{4,}$")


def is_valid_username(username: str, /) -> bool:
    return bool(USERNAME_PATTERN.match(username))


def is_valid_deeplink_parameter(param: str, /) -> bool:
    return bool(DEEPLINK_PAYLOAD_PATTERN.match(param))


def extract_bot_id(token: str) -> int:
    try:
        return int(token.split(":")[0])
    except ValueError:
        raise InvalidTokenError() from None


def ensure_unique(objs: Sequence[T]) -> list[T]:
    tmp: list[T] = []

    for obj in objs:
        if obj not in tmp:
            tmp.append(obj)
    return tmp



from __future__ import annotations

from typing import final

from .abc import TelegramType
from .chat import Chat
from .chat_boost import ChatBoost


@final
class ChatBoostUpdated(TelegramType):
    """This object represents a boost added to a chat or changed.

    See: https://core.telegram.org/bots/api#chatboostupdated
    """

    chat: Chat
    """Chat which was boosted."""

    boost: ChatBoost
    """Infomation about the chat boost."""



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import final

from yatbaf.types import MessageId

from .abc import TelegramMethod

if TYPE_CHECKING:
    from yatbaf.enums import ParseMode
    from yatbaf.types import MessageEntity
    from yatbaf.types import ReplyParameters
    from yatbaf.typing import NoneBool
    from yatbaf.typing import NoneInt
    from yatbaf.typing import ReplyMarkup


@final
class CopyMessage(TelegramMethod[MessageId]):
    """See :meth:`yatbaf.bot.Bot.copy_message`"""

    chat_id: str | int
    from_chat_id: str | int
    message_id: int
    caption: str | None
    message_thread_id: NoneInt = None
    parse_mode: ParseMode | None = None
    caption_entities: list[MessageEntity] | None = None
    disable_notification: NoneBool = None
    protect_content: NoneBool = None
    reply_parameters: ReplyParameters | None = None
    reply_markup: ReplyMarkup | None = None



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import ClassVar
from typing import final

from .abc import TelegramMethodWithMedia

if TYPE_CHECKING:
    from yatbaf.types import InputSticker


@final
class AddStickerToSet(TelegramMethodWithMedia[bool]):
    """See :meth:`yatbaf.bot.Bot.add_sticker_to_set`"""

    user_id: int
    name: str
    sticker: InputSticker

    __meth_media_fields__: ClassVar[tuple[str, ...]] = ("sticker",)



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import final

from .abc import TelegramType
from .callback_query import CallbackQuery
from .chat_boost_removed import ChatBoostRemoved
from .chat_boost_updated import ChatBoostUpdated
from .chat_join_request import ChatJoinRequest
from .chat_member import ChatMemberUpdated
from .chosen_inline_result import ChosenInlineResult
from .inline_query import InlineQuery
from .message import Message
from .message_reaction_count_updated import MessageReactionCountUpdated
from .message_reaction_updated import MessageReactionUpdated
from .poll import Poll
from .poll_answer import PollAnswer
from .pre_checkout_query import PreCheckoutQuery
from .shipping_query import ShippingQuery

if TYPE_CHECKING:
    from yatbaf.typing import UpdateModel


@final
class Update(TelegramType):
    """This object represents an incoming update.

    .. note::

        At most **one** of the optional parameters can be present in any given
        update.

    See: https://core.telegram.org/bots/api#update
    """

    update_id: int
    """The update's unique identifier. Update identifiers start from a certain
    positive number and increase sequentially. This ID becomes especially handy
    if you're using webhooks, since it allows you to ignore repeated updates or
    to restore the correct update sequence, should they get out of order. If
    there are no new updates for at least a week, then identifier of the next
    update will be chosen randomly instead of sequentially.
    """

    message: Message | None = None
    """*Optional.* New incoming message of any kind - text, photo, sticker, etc."""  # noqa: E501

    edited_message: Message | None = None
    """*Optional.* New version of a message that is known to the bot and was edited."""  # noqa: E501

    channel_post: Message | None = None
    """*Optional.* New incoming channel post of any kind - text, photo, sticker, etc."""  # noqa: E501

    edited_channel_post: Message | None = None
    """*Optional.* New version of a channel post that is known to the bot and
    was edited.
    """

    message_reaction: MessageReactionUpdated | None = None
    """*Optional.* A reaction to a message was changed by a user.

    .. important::

        The bot must be an administrator in the chat and must explicitly specify
        ``message_reaction`` in the list of ``allowed_updates`` to receive these
        updates. The update isn't received for reactions set by bots.
    """

    message_reaction_count: MessageReactionCountUpdated | None = None
    """*Optional.* Reactions to a message with anonymous reactions were changed.

    .. important::

        The bot must be an administrator in the chat and must explicitly specify
        ``message_reaction_count`` in the list of ``allowed_updates`` to receive
        these updates.
    """

    inline_query: InlineQuery | None = None
    """*Optional.* New incoming inline query"""

    chosen_inline_result: ChosenInlineResult | None = None
    """*Optional.* The result of an inline query that was chosen by a user and
    sent to their chat partner. Please see documentation on the
    `feedback collecting`_ for details on how to enable these updates for your
    bot.

    .. _feedback collecting: https://core.telegram.org/bots/inline#collecting-feedback
    """  # noqa: E501

    callback_query: CallbackQuery | None = None
    """*Optional.* New incoming callback query."""

    shipping_query: ShippingQuery | None = None
    """*Optional.* New incoming shipping query.

    .. note::

        Only for invoices with flexible price.
    """

    pre_checkout_query: PreCheckoutQuery | None = None
    """*Optional.* New incoming pre-checkout query. Contains full information
    about checkout.
    """

    poll: Poll | None = None
    """*Optional.* New poll state. Bots receive only updates about stopped polls
    and polls, which are sent by the bot.
    """

    poll_answer: PollAnswer | None = None
    """*Optional.* A user changed their answer in a non-anonymous poll. Bots
    receive new votes only in polls that were sent by the bot itself.
    """

    my_chat_member: ChatMemberUpdated | None = None
    """*Optional.* The bot's chat member status was updated in a chat. For
    private chats, this update is received only when the bot is blocked or
    unblocked by the user.
    """

    chat_member: ChatMemberUpdated | None = None
    """*Optional.* A chat member's status was updated in a chat.

    .. important::

        The bot must be an administrator in the chat and must explicitly specify
        ``chat_member`` in the list of ``allowed_updates`` to receive these
        updates.
    """

    chat_join_request: ChatJoinRequest | None = None
    """*Optional.* A request to join the chat has been sent.

    .. important::

        The bot must have the ``can_invite_users`` administrator right in the
        chat to receive these updates.
    """

    chat_boost: ChatBoostUpdated | None = None
    """*Optional.* A chat boost was added or changed.

    .. important::

        The bot must be an administrator in the chat to receive these updates.
    """

    removed_chat_boost: ChatBoostRemoved | None = None
    """*Optional.* A boost was removed from a chat.

    .. important::

        The bot must be an administrator in the chat to receive these updates.
    """

    def filter(self) -> tuple[UpdateModel, str]:  # type: ignore[return]
        """:meta private:"""

        id = self.update_id
        self.update_id = None  # type: ignore[assignment]
        for field in self.__struct_fields__:
            if (v := getattr(self, field)) is not None:
                self.update_id = id
                # one filed is always not None
                return (v, field)



from __future__ import annotations

from typing import final

from .abc import TelegramType


@final
class BotDescription(TelegramType):
    """This object represents the bot's description.

    See: https://core.telegram.org/bots/api#botdescription
    """

    description: str
    """The bot's description"""



from __future__ import annotations

from typing import final

from .abc import TelegramType
from .user import User


@final
class GameHighScore(TelegramType):
    """This object represents one row of the high scores table for a game.

    See: https://core.telegram.org/bots/api#gamehighscore
    """

    position: int
    """Position in high score table for the game."""

    user: User
    """User."""

    score: int
    """Score."""



from __future__ import annotations

from typing import final

from yatbaf.typing import NoneInt

from .abc import TelegramType


@final
class Location(TelegramType):
    """This object represents a point on the map.

    See: https://core.telegram.org/bots/api#location
    """

    longitude: float
    """Longitude as defined by sender."""

    latitude: float
    """Latitude as defined by sender."""

    horizontal_accuracy: float | None = None
    """*Optional.* The radius of uncertainty for the location, measured in
    meters; 0-1500.
    """

    live_period: NoneInt = None
    """*Optional.* Time relative to the message sending date, during which the
    location can be updated; in seconds.

    .. note::

        For active live locations only.
    """

    heading: NoneInt = None
    """*Optional.* The direction in which user is moving, in degrees; 1-360.

    .. note::

        For active live locations only.
    """

    proximity_alert_radius: NoneInt = None
    """*Optional.* The maximum distance for proximity alerts about approaching
    another chat member, in meters.

    .. note::

        For sent live locations only.
    """



from __future__ import annotations

from typing import final

from yatbaf.typing import NoneBool
from yatbaf.typing import NoneStr

from .abc import TelegramType
from .callback_game import CallbackGame
from .login_url import LoginUrl
from .switch_inline_query_chosen_chat import SwitchInlineQueryChosenChat
from .web_app_info import WebAppInfo


@final
class InlineKeyboardButton(TelegramType):
    """This object represents one button of an inline keyboard.

    .. note::

        You **must** use exactly one of the optional fields.

    See: https://core.telegram.org/bots/api#inlinekeyboardbutton
    """

    text: str
    """Label text on the button."""

    url: NoneStr = None
    """*Optional.* HTTP or tg:// URL to be opened when the button is pressed.
    Links tg://user?id=<user_id> can be used to mention a user by their ID
    without using a username, if this is allowed by their privacy settings.
    """

    callback_data: NoneStr = None
    """*Optional.* Data to be sent in a callback query to the bot when button
    is pressed, 1-64 bytes.
    """

    web_app: WebAppInfo | None = None
    """*Optional.* Description of the Web App that will be launched when the
    user presses the button. The Web App will be able to send an arbitrary
    message on behalf of the user using the method
    :meth:`yatbaf.bot.Bot.answer_web_app_query()`.

    .. note::

        Available only in private chats between a user and the bot.
    """

    login_url: LoginUrl | None = None
    """*Optional.* An HTTPS URL used to automatically authorize the user."""

    switch_inline_query: NoneStr = None
    """*Optional.* If set, pressing the button will prompt the user to select
    one of their chats, open that chat and insert the bot's username and the
    specified inline query in the input field. May be empty, in which case just
    the bot's username will be inserted.
    """

    switch_inline_query_current_chat: NoneStr = None
    """*Optional.* If set, pressing the button will insert the bot's username
    and the specified inline query in the current chat's input field. May be
    empty, in which case only the bot's username will be inserted.
    """

    switch_inline_query_chosen_chat: SwitchInlineQueryChosenChat | None = None
    """*Optional.* If set, pressing the button will prompt the user to select
    one of their chats of the specified type, open that chat and insert the
    bot's username and the specified inline query in the input field.
    """

    callback_game: CallbackGame | None = None
    """*Optional.* Description of the game that will be launched when the user
    presses the button.

    .. important::

        This type of button *must* always be the first button in the first row.
    """

    pay: NoneBool = None
    """*Optional.* Specify True, to send a Pay button.

    .. important::

        This type of button must always be the first button in the first row
        and can only be used in invoice messages.
    """



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import final

from yatbaf.types import Message

from .abc import TelegramMethod

if TYPE_CHECKING:
    from yatbaf.enums import ParseMode
    from yatbaf.types import LinkPreviewOptions
    from yatbaf.types import MessageEntity
    from yatbaf.types import ReplyParameters
    from yatbaf.typing import NoneBool
    from yatbaf.typing import NoneInt
    from yatbaf.typing import ReplyMarkup


@final
class SendMessage(TelegramMethod[Message]):
    """See :meth:`yatbaf.bot.Bot.send_message`"""

    chat_id: str | int
    text: str
    message_thread_id: NoneInt = None
    parse_mode: ParseMode | None = None
    entities: list[MessageEntity] | None = None
    link_preview_options: LinkPreviewOptions | None = None
    disable_notification: NoneBool = None
    protect_content: NoneBool = None
    reply_parameters: ReplyParameters | None = None
    reply_markup: ReplyMarkup | None = None



from __future__ import annotations

__all__ = (
    "TelegramMethod",
    "TelegramMethodWithFile",
    "TelegramMethodWithMedia",
)

from collections.abc import Iterable
from secrets import token_urlsafe
from typing import TYPE_CHECKING
from typing import Any
from typing import ClassVar
from typing import Generic
from typing import cast
from typing import get_args

from msgspec import Struct
from msgspec import json as jsonlib
from msgspec import to_builtins

from yatbaf.typing import ResultModelT

if TYPE_CHECKING:
    from typing import TypeAlias

    from yatbaf.types.abc import TelegramType
    from yatbaf.typing import InputFile

Data: TypeAlias = "bytes | dict[str, Any]"
File: TypeAlias = "dict[str, Any]"
Encoded: TypeAlias = "tuple[Data | None, File | None]"

_json_encoder = jsonlib.Encoder()
_file_name_length = 12


class TelegramMethod(Struct, Generic[ResultModelT], omit_defaults=True):
    """Base class for methods"""

    __meth_name__: ClassVar[str]
    __meth_result_model__: ClassVar[Any]

    def __init_subclass__(cls, **kwargs: Any) -> None:
        super().__init_subclass__(**kwargs)

        # Skip for `Telegram*` subclasses
        if cls.__name__.startswith("Telegram"):
            return

        cls.__meth_name__ = cls.__name__.lower()
        cls.__meth_result_model__ = (
            get_args(cls.__orig_bases__[0])[0]  # type: ignore[attr-defined]
        )

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}>"

    def __str__(self) -> str:
        return self.__meth_name__

    @classmethod
    def _get_name(cls) -> str:
        return cls.__meth_name__

    @classmethod
    def _get_result_model(cls) -> type[ResultModelT]:
        return cast("type[ResultModelT]", cls.__meth_result_model__)

    def _encode_params(self) -> Encoded:
        json = j if (j := _json_encoder.encode(self)) != b"{}" else None
        return (json, None)


class TelegramMethodWithFile(TelegramMethod, Generic[ResultModelT]):
    """Base class for methods with file field"""

    __meth_file_fields__: ClassVar[tuple[str, ...]]

    def _encode_params(self) -> Encoded:
        files: dict[str, InputFile] = {}
        for file_field in self.__meth_file_fields__:
            file_obj = getattr(self, file_field)
            if file_obj is None or isinstance(file_obj, str):
                continue

            file_name = token_urlsafe(_file_name_length)
            files[file_name] = file_obj
            setattr(self, file_field, f"attach://{file_name}")

        # no files or it's `file_id`
        if not files:
            return super()._encode_params()

        return (to_builtins(self), files)


class TelegramMethodWithMedia(TelegramMethod, Generic[ResultModelT]):
    """Base class for methods with media field"""

    __meth_media_fields__: ClassVar[tuple[str, ...]]

    def _encode_params(self) -> Encoded:
        files: dict[str, InputFile] = {}

        for media_field in self.__meth_media_fields__:
            content: TelegramType | list[TelegramType] = getattr(
                self, media_field
            )
            media = content if isinstance(content, Iterable) else (content,)
            for media_obj in media:
                for file_field in media_obj.__type_file_fields__:
                    file_obj = getattr(media_obj, file_field)
                    if file_obj is None or isinstance(file_obj, str):
                        continue

                    file_name = token_urlsafe(_file_name_length)
                    files[file_name] = file_obj
                    setattr(media_obj, file_field, f"attach://{file_name}")

        if not files:
            return super()._encode_params()

        for media_field in self.__meth_media_fields__:
            setattr(
                self,
                media_field,
                _json_encoder.encode(getattr(self, media_field)).decode(),
            )

        return (to_builtins(self), files)



from __future__ import annotations

from typing import final

from .abc import TelegramMethod


@final
class ExportChatInviteLink(TelegramMethod[str]):
    """See :meth:`yatbaf.bot.Bot.export_chat_invite_link`"""

    chat_id: str | int



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import final

from yatbaf.types import Message

from .abc import TelegramMethod

if TYPE_CHECKING:
    from yatbaf.types import InlineKeyboardMarkup
    from yatbaf.types import LabeledPrice
    from yatbaf.types import ReplyParameters
    from yatbaf.typing import NoneBool
    from yatbaf.typing import NoneInt
    from yatbaf.typing import NoneStr


@final
class SendInvoice(TelegramMethod[Message]):
    """See :meth:`yatbaf.bot.Bot.send_invoice`"""

    chat_id: str | int
    title: str
    description: str
    payload: str
    provider_token: str
    currency: str
    prices: list[LabeledPrice]
    message_thread_id: NoneInt = None
    max_tip_amount: NoneInt = None
    suggested_tip_amounts: list[int] | None = None
    start_parameter: NoneStr = None
    provider_data: NoneStr = None
    photo_url: str | None = None
    photo_size: NoneInt = None
    photo_width: NoneInt = None
    photo_height: NoneInt = None
    need_name: NoneBool = None
    need_phone_number: NoneBool = None
    need_email: NoneBool = None
    need_shipping_address: NoneBool = None
    send_phone_number_to_provider: NoneBool = None
    send_email_to_provider: NoneBool = None
    is_flexible: NoneBool = None
    disable_notification: NoneBool = None
    protect_content: NoneBool = None
    reply_parameters: ReplyParameters | None = None
    reply_markup: InlineKeyboardMarkup | None = None



from __future__ import annotations

__all__ = ("Command",)

from typing import TYPE_CHECKING
from typing import final

from yatbaf.utils import parse_command

if TYPE_CHECKING:
    from yatbaf.types import Message


@final
class Command:
    """Command filter.

    Usage::

        @on_message(filters=[Command("hello", "start")])
        async def cmd_start(message: Message) -> None:
            ...

        @on_message(filters=[Command("ping")])
        async def cmd_ping(message: Message) -> None:
            ...
    """

    __slots__ = (
        "commands",
        "priority",
    )

    def __init__(self, *command: str, priority: int = 400) -> None:
        """
        :param command: The commnad or list of commands.
        :param priority: Filter priority. Default 400.
        :raise ValueError: If ``command`` is empty.

        .. note::

            ``command`` case-insensitive.
        """

        if not command:
            raise ValueError("You must pass at least one command.")

        self.commands = frozenset([
            c.removeprefix("/").lower() for c in command
        ])
        self.priority = priority

    def check(self, update: Message) -> bool:
        if (text := update.text) is None:
            return False

        return ((command := parse_command(text)) is not None
                and command in self.commands)



from __future__ import annotations

from typing import final

from yatbaf.typing import NoneBool
from yatbaf.typing import NoneStr

from .abc import TelegramType


@final
class ForceReply(TelegramType):
    """Upon receiving a message with this object, Telegram clients will display
    a reply interface to the user (act as if the user has selected the bot's
    message and tapped 'Reply').

    See: https://core.telegram.org/bots/api#forcereply
    https://core.telegram.org/bots/features#privacy-mode
    """

    force_reply: bool
    """Shows reply interface to the user, as if they manually selected the
    bot's message and tapped 'Reply'.
    """

    input_field_placeholder: NoneStr = None
    """*Optional.* The placeholder to be shown in the input field when the reply
    is active.

    .. note::

        1-64 characters.
    """

    selective: NoneBool = None
    """*Optional.* Use this parameter if you want to force reply from specific
    users only. Targets: 1) users that are ``@mentioned`` in the text of the
    :class:`yatbaf.types.Message` object; 2) if the bot's message is a reply
    (has ``reply_to_message_id``), sender of the original message.
    """



import pytest

from yatbaf.dispatcher import Dispatcher


def test_dispatcher(token):
    dispatcher = Dispatcher(token)
    assert dispatcher._routers
    assert not dispatcher._middleware
    assert not dispatcher._guards


def test_update_type(token):
    with pytest.raises(RuntimeError):
        Dispatcher(token)._update_type



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import final

from .abc import TelegramMethod

if TYPE_CHECKING:
    from yatbaf.types import PassportElementError


@final
class SetPassportDataErrors(TelegramMethod[bool]):
    """See :meth:`yatbaf.bot.Bot.set_passport_data_errors`"""

    user_id: int
    errors: list[PassportElementError]



from __future__ import annotations

__all__ = ("Dispatcher",)

import logging
from collections import defaultdict
from typing import TYPE_CHECKING
from typing import Any

from .background import BackgroundTasks
from .enums import BotEnvi
from .error_handler import DefaultErrorHandler
from .models import UpdateInfo
from .router import BaseRouter
from .router import router_map
from .telegram import TelegramClient
from .utils import decode_webhook
from .utils import extract_bot_id
from .warnings import warn_duplicate

if TYPE_CHECKING:
    from collections.abc import Sequence

    from .abc import AbstractClient
    from .abc import AbstractRouter
    from .handler import Handler
    from .methods.abc import TelegramMethod
    from .middleware import Middleware
    from .types import Update
    from .typing import ErrorHandler
    from .typing import GuardCallable
    from .typing import HandlerMiddleware
    from .typing import ResultModelT

log = logging.getLogger(__name__)


class Dispatcher(BaseRouter):
    """Dispatcher"""

    __slots__ = (
        "_api_client",
        "_error_handler",
        "_tasks",
        "_id",
        "_routers",
        "_ctx",
        "_middleware_stack",
    )

    def __init__(
        self,
        token: str,
        handlers: Sequence[Handler] | None = None,
        routers: Sequence[AbstractRouter] | None = None,
        middleware: Sequence[Middleware | HandlerMiddleware] | None = None,
        guards: Sequence[GuardCallable] | None = None,
        error_handler: ErrorHandler | None = None,
        api_url: str | None = None,
        environment: BotEnvi | None = None,
        client: AbstractClient | None = None,
    ) -> None:
        """
        :param middleware: *Optional.* A sequence of :class:`~yatbaf.middleware.Middleware`
            or :class:`~yatbaf.typing.HandlerMiddleware`.
        :param guards: *Optional.* A sequence of :class:`~yatbaf.typing.Guard`.
        :param handlers: *Optional.* A sequence of :class:`~yatbaf.handler.Handler`.
        :param routers: *Optional.* A sequence of :class:`~yatbaf.abc.AbstractRouter`.
        :param api_url: *Optional.* Api server address. Defautl to :attr:`~yatbaf.telegram.SERVER_URL`.
        :param environment: *Optional.* Bot environment (see :class:`~yatbaf.enums.BotEnvi`).
        :param client: *Optional.* Http client.
        """  # noqa: E501
        self._api_client = TelegramClient(
            token=token,
            api_url=api_url,
            environment=environment,
            client=client,
        )
        self._error_handler = DefaultErrorHandler(error_handler)
        self._tasks = BackgroundTasks(self._error_handler)
        self._routers: dict[str, list[AbstractRouter]] = {
            "message": [],
            "edited_message": [],
            "channel_post": [],
            "edited_channel_post": [],
            "message_reaction": [],
            "message_reaction_count": [],
            "inline_query": [],
            "chosen_inline_result": [],
            "callback_query": [],
            "shipping_query": [],
            "pre_checkout_query": [],
            "poll": [],
            "poll_answer": [],
            "my_chat_member": [],
            "chat_member": [],
            "chat_join_request": [],
            "chat_boost": [],
            "removed_chat_boost": [],
        }
        self._id = extract_bot_id(token)
        self._ctx: dict[str, Any] = {}
        self._middleware_stack = self.__resolve

        super().__init__(
            routers=routers,
            handlers=handlers,
            middleware=middleware,
            guards=guards,
        )
        self._on_registration()

    @property
    def id(self) -> int:  # noqa: A003
        """Bot ID"""
        return self._id

    @property
    def ctx(self) -> dict[str, Any]:
        """Dict object.

        Use it to store extra state on bot instance.
        """
        return self._ctx

    @property
    def tasks(self) -> BackgroundTasks:
        """Task manager."""
        return self._tasks

    @property
    def _update_type(self) -> str:
        raise RuntimeError(
            "Dispatcher can't be used here. "
            "You are trying to do something illegal ._."
        )

    def _register_routers(self, routers: Sequence[AbstractRouter]) -> None:
        for router in routers:
            parent = router._parent
            if parent is not None and parent is not self:
                raise ValueError(f"{router!r} already registered in {parent!r}")

            router_list = self._routers[router._update_type]
            if parent or router in router_list:
                warn_duplicate(router, self)
                continue

            router_list.append(router)
            router._parent = self

    def _register_handlers(self, handlers: Sequence[Handler]) -> None:
        routers: dict[str, list[Handler]] = defaultdict(list)
        for handler in handlers:
            routers[handler.update_type].append(handler)

        self._register_routers([
            router_map[t](handlers=h) for t, h in routers.items()
        ])

    def _bind_self(self, update: ResultModelT, /) -> ResultModelT:
        raise NotImplementedError()

    def _on_registration(self) -> None:
        self._register_middleware(self._middleware)
        for routers in self._routers.values():
            for router in routers:
                router._on_registration()

    async def _call_api(
        self, method: TelegramMethod[ResultModelT]
    ) -> ResultModelT:
        """:meta private:"""
        return self._bind_self((await self._api_client.invoke(method)).result)

    async def __resolve(self, update: UpdateInfo, /) -> bool:
        """:meta private:"""
        content = update.content
        if await self._check_guards(content):
            routers = self._routers[update.name]
            for router in routers:
                if await router._resolve(update):
                    return True
        return False

    async def _resolve(self, update: UpdateInfo, /) -> bool:
        return await self._middleware_stack(update)

    async def process_update(self, data: Update | bytes, /) -> None:
        """Process incoming update.

        :param data: :class:`~yatbaf.types.update.Update` instance (long
            polling) or :class:`bytes` (webhook content).
        """
        # catch any exceptions in guard/middleware/handler
        try:
            if isinstance(data, bytes):
                data = decode_webhook(data)

            content, name = data.filter()
            await self._resolve(
                UpdateInfo(
                    id=data.update_id,
                    name=name,
                    content=self._bind_self(content),
                )
            )

        except Exception as error:
            await self._error_handler.on_error(error)

    async def shutdown(self, force: bool = False) -> None:
        """Cleanup resources.

        :param force: Pass ``True`` if you want to cancel all active tasks
            stored in :attr:`.tasks`.
        """
        log.debug("bot shutting down...")
        if not force:
            # wait for active tasks
            await self._tasks.wait_all()
        else:
            await self._tasks.kill_all()
        # close http client
        await self._api_client.close()



from __future__ import annotations

__all__ = (
    "ReactionType",
    "ReactionTypeEmoji",
    "ReactionTypeCustomEmoji",
)

from typing import ClassVar
from typing import Literal
from typing import TypeAlias
from typing import final

from .abc import TelegramType


@final
class ReactionTypeEmoji(TelegramType, tag="emoji"):
    """The reaction is based on an emoji.

    See: https://core.telegram.org/bots/api#reactiontypeemoji
    """

    emoji: str
    """Reaction emoji."""

    type: ClassVar[Literal["emoji"]] = "emoji"
    """Type of the reaction, always *emoji*"""


@final
class ReactionTypeCustomEmoji(TelegramType, tag="custom_emoji"):
    """The reaction is based on a custom emoji.

    See: https://core.telegram.org/bots/api#reactiontypecustomemoji
    """

    custom_emoji_id: str
    """Custom emoji identifier."""

    type: ClassVar[Literal["custom_emoji"]] = "custom_emoji"
    """Type of the reaction, always *custom_emoji*."""


ReactionType: TypeAlias = ReactionTypeEmoji | ReactionTypeCustomEmoji
"""This object describes the type of a reaction.

See: https://core.telegram.org/bots/api#reactiontype
"""



from __future__ import annotations

from typing import final

from .abc import TelegramMethod


@final
class GetChatMemberCount(TelegramMethod[int]):
    """See :meth:`yatbaf.bot.Bot.get_chat_member_count`"""

    chat_id: str | int



from __future__ import annotations

from typing import final

from yatbaf.typing import NoneBool

from .abc import TelegramType


@final
class ChatAdministratorRights(TelegramType):
    """Represents the rights of an administrator in a chat.

    See: https://core.telegram.org/bots/api#chatadministratorrights
    """

    is_anonymous: bool
    """``True``, if the user's presence in the chat is hidden."""

    can_manage_chat: bool
    """``True``, if the administrator can access the chat event log, chat
    statistics, message statistics in channels, see channel members, see
    anonymous administrators in supergroups and ignore slow mode. Implied by
    any other administrator privilege.
    """

    can_delete_messages: bool
    """``True``, if the administrator can delete messages of other users."""

    can_manage_video_chats: bool
    """``True``, if the administrator can manage video chats."""

    can_restrict_members: bool
    """``True``, if the administrator can restrict, ban or unban chat members."""  # noqa: E501

    can_promote_members: bool
    """``True``, if the administrator can add new administrators with a subset
    of their own privileges or demote administrators that they have promoted,
    directly or indirectly (promoted by administrators that were appointed by
    the user).
    """

    can_change_info: bool
    """``True``, if the user is allowed to change the chat title, photo and
    other settings.
    """

    can_invite_users: bool
    """``True``, if the user is allowed to invite new users to the chat."""

    can_post_messages: NoneBool = None
    """*Optional.* ``True``, if the administrator can post in the channel.

    .. note::

        Channels only.
    """

    can_edit_messages: NoneBool = None
    """*Optional.* ``True``, if the administrator can edit messages of other
    users and can pin messages.

    .. note::

        Channels only.
    """

    can_pin_messages: NoneBool = None
    """*Optional.* ``True``, if the user is allowed to pin messages.

    .. note::

        Groups and supergroups only.
    """

    can_post_stories: NoneBool = None
    """*Optional.* ``True``, if the administrator can post stories in the
    channel.

    .. note::

        Channels only.
    """

    can_edit_stories: NoneBool = None
    """Optional. ``True``, if the administrator can edit stories posted by other
    users.

    .. note::

        Channels only.
    """

    can_delete_stories: NoneBool = None
    """*Optional.* ``True``, if the administrator can delete stories posted by
    other users.

    .. note::

        Channels only.
    """

    can_manage_topics: NoneBool = None
    """*Optional.* ``True``, if the user is allowed to create, rename, close,
    and reopen forum topics.

    .. note::

        Supergroups only.
    """



from __future__ import annotations

from typing import final

from yatbaf.typing import NoneStr

from .abc import TelegramType


@final
class ForumTopic(TelegramType):
    """This object represents a forum topic.

    See: https://core.telegram.org/bots/api#forumtopic
    """

    message_thread_id: int
    """Unique identifier of the forum topic."""

    name: str
    """Name of the topic."""

    icon_color: int
    """Color of the topic icon in RGB format."""

    icon_custom_emoji_id: NoneStr = None
    """
    *Optional.* Unique identifier of the custom emoji shown as the topic icon.
    """



from __future__ import annotations

__all__ = (
    "BotCommandScopeAllChatAdministrators",
    "BotCommandScopeAllGroupChats",
    "BotCommandScopeAllPrivateChats",
    "BotCommandScopeChat",
    "BotCommandScopeChatAdministrators",
    "BotCommandScopeChatMember",
    "BotCommandScopeDefault",
    "BotCommandScope",
)

from typing import Literal
from typing import TypeAlias
from typing import final

from msgspec import field

from .abc import TelegramType


@final
class BotCommandScopeAllChatAdministrators(TelegramType):
    """Represents the scope of bot commands, covering all group and supergroup
    chat administrators.

    See: https://core.telegram.org/bots/api#botcommandscopeallchatadministrators
    """

    type: Literal["all_chat_administrators"] = field(
        default_factory=lambda: "all_chat_administrators"
    )
    """Scope type, must be `all_chat_administrators`."""


@final
class BotCommandScopeAllGroupChats(TelegramType):
    """Represents the scope of bot commands, covering all group and supergroup
    chats.

    See: https://core.telegram.org/bots/api#botcommandscopeallgroupchats
    """

    type: Literal["all_group_chats"] = field(
        default_factory=lambda: "all_group_chats"
    )
    """Scope type, must be `all_group_chats`."""


@final
class BotCommandScopeAllPrivateChats(TelegramType):
    """Represents the scope of bot commands, covering all private chats.

    See: https://core.telegram.org/bots/api#botcommandscopeallprivatechats
    """

    type: Literal["all_private_chats"] = field(
        default_factory=lambda: "all_private_chats"
    )
    """Scope type, must be `all_private_chats`."""


@final
class BotCommandScopeChat(TelegramType):
    """Represents the scope of bot commands, covering a specific chat.

    See: https://core.telegram.org/bots/api#botcommandscopechat
    """

    chat_id: int | str
    """Unique identifier for the target chat or username of the target
    supergroup (in the format @supergroupusername).
    """

    type: Literal["chat"] = field(default_factory=lambda: "chat")
    """Scope type, must be `chat`."""


@final
class BotCommandScopeChatAdministrators(TelegramType):
    """Represents the scope of bot commands, covering all administrators of a
    specific group or supergroup chat.

    See: https://core.telegram.org/bots/api#botcommandscopechatadministrators
    """

    chat_id: int | str
    """Unique identifier for the target chat or username of the target
    supergroup (in the format @supergroupusername).
    """

    type: Literal["chat_administrators"] = field(
        default_factory=lambda: "chat_administrators"
    )
    """Scope type, must be `chat_administrators`."""


@final
class BotCommandScopeChatMember(TelegramType):
    """Represents the scope of bot commands, covering a specific member of a
    group or supergroup chat.

    See: https://core.telegram.org/bots/api#botcommandscopechatmember
    """

    chat_id: int | str
    """Unique identifier for the target chat or username of the target
    supergroup (in the format @supergroupusername).
    """

    user_id: int
    """Unique identifier of the target user."""

    type: Literal["chat_member"] = field(default_factory=lambda: "chat_member")
    """Scope type, must be `chat_member`."""


@final
class BotCommandScopeDefault(TelegramType):
    """Represents the default scope of bot commands. Default commands are used
    if no commands with a narrower scope are specified for the user.

    See: https://core.telegram.org/bots/api#botcommandscopedefault
    """

    type: Literal["default"] = field(default_factory=lambda: "default")
    """Scope type, must be `default`."""


# https://core.telegram.org/bots/api#botcommandscope
BotCommandScope: TypeAlias = (
    "BotCommandScopeAllChatAdministrators "
    "| BotCommandScopeAllGroupChats "
    "| BotCommandScopeAllPrivateChats "
    "| BotCommandScopeChat "
    "| BotCommandScopeChatAdministrators "
    "| BotCommandScopeChatMember "
    "| BotCommandScopeDefault"
)



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import TypeAlias

if TYPE_CHECKING:
    from .inaccessible_message import InaccessibleMessage
    from .message import Message

MaybeInaccessibleMessage: TypeAlias = "Message | InaccessibleMessage"
"""This object describes a message that can be inaccessible to the bot.

See: https://core.telegram.org/bots/api#maybeinaccessiblemessage
"""



from __future__ import annotations

__all__ = ("HttpClient",)

import logging
from contextlib import asynccontextmanager
from typing import TYPE_CHECKING
from typing import final

import httpx

from .abc import AbstractClient
from .exceptions import RequestTimeoutError
from .models import HTTPResponse
from .models import Request

if TYPE_CHECKING:
    from collections.abc import AsyncIterator
    from typing import Final

log = logging.getLogger(__name__)


@final
class HttpClient(AbstractClient):
    """Default http client."""

    __slots__ = ("_client",)

    def __init__(
        self,
        *,
        timeout: float = 8.0,
        connect_timeout: float = 16.0,
    ) -> None:
        """
        :param timeout: *Optional.* Request read/write timeout.
        :param connect_timeout: *Optional.* Request connect timeout.
        """
        self._client: Final[httpx.AsyncClient] = httpx.AsyncClient(
            timeout=httpx.Timeout(
                timeout=timeout,
                connect=connect_timeout,
            ),
        )

    async def send_post(
        self,
        request: Request, /,
        timeout: float | None = None,
    ) -> HTTPResponse[bytes]:  # yapf: disable
        try:
            response = await self._client.post(
                url=request.url,
                content=request.content,
                headers=request.headers,
                data=request.data,
                files=request.files,
                timeout=httpx.Timeout(
                    timeout=timeout,
                    connect=self._client.timeout.connect,
                ) if timeout is not None else httpx.USE_CLIENT_DEFAULT,
            )
        except httpx.TimeoutException as error:
            raise RequestTimeoutError(
                method=request.method,
                orig=error,
            ) from None

        return HTTPResponse(
            status_code=response.status_code,
            content=response.content,
        )

    @asynccontextmanager
    async def file_stream(
        self, url: str, chunk_size: int
    ) -> AsyncIterator[HTTPResponse[AsyncIterator[bytes]]]:
        try:
            async with self._client.stream("GET", url) as r:
                response = HTTPResponse(
                    status_code=r.status_code,
                    content=r.aiter_bytes(chunk_size),
                )
                yield response
        except httpx.TimeoutException as error:
            raise RequestTimeoutError(
                method="file",
                orig=error,
            ) from None

    async def close(self) -> None:
        log.debug("Shutting down")
        await self._client.aclose()



from __future__ import annotations

__all__ = (
    "Chat",
    "Group",
    "Channel",
)

from typing import TYPE_CHECKING
from typing import final

from yatbaf.enums import ChatType

if TYPE_CHECKING:
    from yatbaf.types import Message


@final
class Chat:
    """Private chat filter.

    Use it to filter messages coming from private chats::

        @on_message(fileters=[Chat()])
        async def handler(message: Message) -> None:
            '''messages from any Private chats'''
    """

    __slots__ = ("priority",)

    def __init__(self, *, priority: int = 100) -> None:
        """
        :param priority: Filter priority. Default 100.
        """

        self.priority = priority

    def check(self, update: Message) -> bool:
        return (
            t := update.chat.type
        ) is ChatType.PRIVATE or t is ChatType.SENDER


@final
class Group:
    """Group or supergroup filter.

    Use it to filter messages coming from Groups or Supergroups::

        @on_message(fileters=[Group()])
        async def handler(message: Message) -> None:
            '''messages from any Group'''

        @on_message(fileters=[Group(1234, 4321)])
        async def specific_groip(message: Message) -> None:
            '''messages from Group with ID 1234 or 4321'''
    """

    __slots__ = (
        "ids",
        "priority",
    )

    def __init__(self, *ids: int, priority: int = 100) -> None:
        """
        :param ids: *Optional.* Group ids.
        :param priority: Filter priority. Default 100.
        """

        self.ids = frozenset(ids)
        self.priority = priority

    def check(self, update: Message) -> bool:
        is_group = (
            t := update.chat.type
        ) is ChatType.GROUP or t is ChatType.SUPERGROUP
        if self.ids:
            return is_group and update.chat.id in self.ids
        return is_group


@final
class Channel:
    """Channel filter.

    Use it to filter messages coming from Channels::

        @on_message(filters=[Channel()])
        async def handler(message: Message) -> None:
            '''messages from any Channel'''

        @on_message(filters=[Channel(1234, 4321)])
        async def handler(message: Message) -> None:
            '''messages from Channel with ID 1234 or 4321'''
    """

    __slots__ = (
        "ids",
        "priority",
    )

    def __init__(self, *ids: int, priority: int = 100) -> None:
        """
        :param ids: *Optional.* Channel ids.
        :param priority: Filter priority. Default 100.
        """

        self.ids = frozenset(ids)
        self.priority = priority

    def check(self, update: Message) -> bool:
        is_channel = update.chat.type is ChatType.CHANNEL
        if self.ids:
            return is_channel and update.chat.id in self.ids
        return is_channel



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import final

from yatbaf.types import Poll

from .abc import TelegramMethod

if TYPE_CHECKING:
    from yatbaf.types import InlineKeyboardMarkup


@final
class StopPoll(TelegramMethod[Poll]):
    """See :meth:`yatbaf.bot.Bot.stop_poll`"""

    chat_id: str | int
    message_id: int
    reply_markup: InlineKeyboardMarkup | None = None



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import ClassVar
from typing import final

from .abc import TelegramMethodWithFile

if TYPE_CHECKING:
    from yatbaf.typing import InputFile


@final
class SetStickerSetThumbnail(TelegramMethodWithFile[bool]):
    """See :meth:`yatbaf.bot.Bot.set_sticker_set_thumbnail`"""

    name: str
    user_id: int
    thumbnail: InputFile | str | None = None

    __meth_file_fields__: ClassVar[tuple[str, ...]] = ("thumbnail",)



from __future__ import annotations

from typing import final

from .abc import TelegramType


@final
class PollOption(TelegramType):
    """This object contains information about one answer option in a poll.

    See: https://core.telegram.org/bots/api#polloption
    """

    text: str
    """Option text, 1-100 characters."""

    voter_count: int
    """Number of users that voted for this option."""



from __future__ import annotations

from typing import final

from .abc import TelegramType
from .labeled_price import LabeledPrice


@final
class ShippingOption(TelegramType):
    """This object represents one shipping option.

    See: https://core.telegram.org/bots/api#shippingoption
    """

    id: str
    """Shipping option identifier."""

    title: str
    """Option title."""

    prices: list[LabeledPrice]
    """List of price portions."""



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import final

from msgspec import field

from yatbaf.enums import ChatType

from .abc import TelegramType
from .location import Location
from .user import User

if TYPE_CHECKING:
    from yatbaf.typing import NoneBool
    from yatbaf.typing import NoneInt
    from yatbaf.typing import NoneStr

    from .inline_query_result import InlineQueryResult
    from .inline_query_result import InlineQueryResultsButton


@final
class InlineQuery(TelegramType):
    """This object represents an incoming inline query. When the user sends an
    empty query, your bot could return some default or trending results.

    See: https://core.telegram.org/bots/api#inlinequery
    """

    id: str
    """Unique identifier for this query."""

    from_: User = field(name="from")
    """Sender"""

    query: str
    """Text of the query (up to 256 characters)"""

    offset: str
    """Offset of the results to be returned, can be controlled by the bot"""

    chat_type: ChatType | None = None
    """*Optional.* Type of the chat from which the inline query was sent.

    .. note::

        The chat type should be always known for requests sent from official
        clients and most third-party clients, unless the request was sent from
        a secret chat.
    """

    location: Location | None = None
    """*Optional.* Sender location, only for bots that request user location."""

    async def answer(
        self,
        results: list[InlineQueryResult],
        *,
        cache_time: NoneInt = None,
        is_personal: NoneBool = None,
        next_offset: NoneStr = None,
        button: InlineQueryResultsButton | None = None,
    ) -> bool:
        """See :meth:`~yatbaf.bot.Bot.answer_inline_query`"""
        return await self.bot.answer_inline_query(
            inline_query_id=self.id,
            results=results,
            cache_time=cache_time,
            is_personal=is_personal,
            next_offset=next_offset,
            button=button,
        )



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import final

from yatbaf.types import Message

from .abc import TelegramMethod

if TYPE_CHECKING:
    from yatbaf.types import ReplyParameters
    from yatbaf.typing import NoneBool
    from yatbaf.typing import NoneInt
    from yatbaf.typing import NoneStr
    from yatbaf.typing import ReplyMarkup


@final
class SendVenue(TelegramMethod[Message]):
    """See :meth:`yatbaf.bot.Bot.send_venue`"""

    chat_id: str | int
    latitude: float
    longitude: float
    title: str
    address: str
    message_thread_id: NoneInt = None
    foursquare_id: NoneStr = None
    foursquare_type: NoneStr = None
    google_place_id: NoneStr = None
    google_place_type: NoneStr = None
    disable_notification: NoneBool = None
    protect_content: NoneBool = None
    reply_parameters: ReplyParameters | None = None
    reply_markup: ReplyMarkup | None = None



from __future__ import annotations

from typing import final

from yatbaf.typing import NoneStr

from .abc import TelegramType


@final
class InputContactMessageContent(TelegramType):
    """Represents the content of a contact message to be sent as the result of
    an inline query.

    See: https://core.telegram.org/bots/api#inputcontactmessagecontent
    """

    phone_number: str
    """Contact's phone number."""

    first_name: str
    """Contact's first name."""

    last_name: NoneStr = None
    """*Optional.* Contact's last name."""

    vcard: NoneStr = None
    """*Optional.* Additional data about the contact in the form of a vCard,
    0-2048 bytes.
    """



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import ClassVar
from typing import final

from yatbaf.types import File

from .abc import TelegramMethodWithFile

if TYPE_CHECKING:
    from yatbaf.enums import StickerFormat
    from yatbaf.typing import InputFile


@final
class UploadStickerFile(TelegramMethodWithFile[File]):
    """See :meth:`yatbaf.bot.Bot.upload_sticker_file`"""

    user_id: int
    sticker: InputFile
    sticker_format: StickerFormat

    __meth_file_fields__: ClassVar[tuple[str, ...]] = ("sticker",)



from __future__ import annotations

from typing import final

from .abc import TelegramType


@final
class WebAppData(TelegramType):
    """Describes data sent from a Web App to the bot.

    See: https://core.telegram.org/bots/api#webappdata
    """

    data: str
    """The data. Be aware that a bad client can send arbitrary data in this field."""  # noqa: E501

    button_text: str
    """Text of the web_app keyboard button from which the Web App was opened.
    Be aware that a bad client can send arbitrary data in this field.
    """



from __future__ import annotations

from typing import final

from .abc import TelegramType


@final
class BotShortDescription(TelegramType):
    """This object represents the bot's short description.

    See: https://core.telegram.org/bots/api#botshortdescription
    """

    short_description: str
    """The bot's short description"""



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import final

from .abc import TelegramType

if TYPE_CHECKING:
    from yatbaf.enums import Currency
    from yatbaf.typing import NoneBool
    from yatbaf.typing import NoneInt
    from yatbaf.typing import NoneStr

    from .labeled_price import LabeledPrice


@final
class InputInvoiceMessageContent(TelegramType):
    """Represents the content of an invoice message to be sent as the result
    of an inline query.

    See: https://core.telegram.org/bots/api#inputinvoicemessagecontent
    """

    title: str
    """Product name, 1-32 characters."""

    description: str
    """Product description, 1-255 characters."""

    payload: str
    """Bot-defined invoice payload, 1-128 bytes. This will not be displayed to
    the user, use for your internal processes.
    """

    provider_token: str
    """Payment provider token, obtained via @BotFather."""

    currency: Currency
    """Three-letter ISO 4217 currency code."""

    prices: list[LabeledPrice]
    """Price breakdown, a list of components."""

    max_tip_amount: NoneInt = None
    """*Optional.* The maximum accepted amount for tips in the smallest units of
    the currency (integer, not float/double).
    """

    suggested_tip_amounts: list[int] | None = None
    """*Optional.* A list of suggested amounts of tip in the smallest units of
    the currency (integer, not float/double).
    """

    provider_data: NoneStr = None
    """*Optional.* A JSON-serialized object for data about the invoice, which
    will be shared with the payment provider. A detailed description of the
    required fields should be provided by the payment provider.
    """

    photo_url: NoneStr = None
    """*Optional.* URL of the product photo for the invoice. Can be a photo of
    the goods or a marketing image for a service.
    """

    photo_size: NoneInt = None
    """*Optional.* Photo size in bytes."""

    photo_width: NoneInt = None
    """*Optional.* Photo width."""

    photo_height: NoneInt = None
    """*Optional.* Photo height."""

    need_name: NoneBool = None
    """*Optional.* Pass ``True`` if you require the user's full name to
    complete the order.
    """

    need_phone_number: NoneBool = None
    """*Optional.* Pass ``True`` if you require the user's phone number to
    complete the order.
    """

    need_email: NoneBool = None
    """*Optional.* Pass ``True`` if you require the user's email address to
    complete the order.
    """

    need_shipping_address: NoneBool = None
    """*Optional.* Pass ``True`` if you require the user's shipping address to
    complete the order.
    """

    send_phone_number_to_provider: NoneBool = None
    """*Optional.* Pass ``True`` if the user's phone number should be sent to
    provider.
    """

    send_email_to_provider: NoneBool = None
    """*Optional.* Pass ``True`` if the user's email address should be sent to
    provider.
    """

    is_flexible: NoneBool = None
    """*Optional.* Pass ``True`` if the final price depends on the shipping
    method.
    """



from __future__ import annotations

from typing import final

from .abc import TelegramType
from .inline_keyboard_button import InlineKeyboardButton


@final
class InlineKeyboardMarkup(TelegramType):
    """This object represents an inline keyboard that appears right next to the
    message it belongs to.

    See: https://core.telegram.org/bots/api#inlinekeyboardmarkup
    """

    inline_keyboard: list[list[InlineKeyboardButton]]
    """Array of button rows."""



import json as jsonlib
from pathlib import Path

import pytest


@pytest.fixture(scope="session")
def api_spec():
    api_json_file = Path(__file__).parent / "data" / "api.json"
    with api_json_file.open("rb") as f:
        return jsonlib.load(f)



from __future__ import annotations

from typing import final

from yatbaf.typing import NoneBool
from yatbaf.typing import NoneStr

from .abc import TelegramType


@final
class LinkPreviewOptions(TelegramType):
    """Describes the options used for link preview generation.

    See: https://core.telegram.org/bots/api#linkpreviewoptions
    """

    is_disabled: NoneBool = None
    """*Optional.* ``True``, if the link preview is disabled."""

    url: NoneStr = None
    """*Optional.* URL to use for the link preview. If empty, then the first URL
    found in the message text will be used.
    """

    prefer_small_media: NoneBool = None
    """*Optional.* ``True``, if the media in the link preview is suppposed to be
    shrunk.

    .. note::

        Ignored if the URL isn't explicitly specified or media size change isn't
        supported for the preview.
    """

    prefer_large_media: NoneBool = None
    """*Optional.* ``True``, if the media in the link preview is suppposed to be
    enlarged.

    .. note::

        Ignored if the URL isn't explicitly specified or media size change isn't
        supported for the preview.
    """

    show_above_text: NoneBool = None
    """*Optional.* ``True``, if the link preview must be shown above the message
    text; otherwise, the link preview will be shown below the message text.
    """



import pytest

from yatbaf.dispatcher import Dispatcher
from yatbaf.enums import UpdateType
from yatbaf.handler import Handler


def test_add_handler(token, handler_func):
    dispatcher = Dispatcher(
        token,
        handlers=[
            Handler(handler_func, UpdateType.MESSAGE),
        ],
    )
    assert dispatcher._routers["message"]
    assert len(dispatcher._routers["message"]) == 1


def test_duplicate_equal_objects(token, handler_func):
    with pytest.warns(UserWarning):
        dispatcher = Dispatcher(
            token,
            handlers=[
                Handler(handler_func, UpdateType.MESSAGE),
                Handler(handler_func, UpdateType.MESSAGE),
            ],
        )
    assert dispatcher._routers["message"]
    assert len(dispatcher._routers["message"]) == 1


def test_duplicate_same_object(token, handler_func):
    handler = Handler(handler_func, UpdateType.MESSAGE)
    with pytest.warns(UserWarning):
        dispatcher = Dispatcher(
            token,
            handlers=[
                handler,
                handler,
            ],
        )
    assert dispatcher._routers["message"]
    assert len(dispatcher._routers["message"]) == 1


def test_order(token):
    dispatcher = Dispatcher(
        token,
        handlers=[
            h1 := Handler(object(), UpdateType.MESSAGE),
            h3 := Handler(object(), UpdateType.MESSAGE),
            h2 := Handler(object(), UpdateType.MESSAGE),
        ]
    )
    assert len(dispatcher._routers["message"]) == 1
    assert len(dispatcher._routers["message"][-1]._handlers) == 3
    assert dispatcher._routers["message"][-1]._handlers[0] is h1
    assert dispatcher._routers["message"][-1]._handlers[1] is h3
    assert dispatcher._routers["message"][-1]._handlers[2] is h2


def test_add_handler_message(token, handler_func):
    dispatcher = Dispatcher(
        token,
        handlers=[
            h := Handler(handler_func, UpdateType.MESSAGE),
        ],
    )
    assert dispatcher._routers["message"][-1]._handlers[0] is h


def test_add_handler_edited_message(token, handler_func):
    dispatcher = Dispatcher(
        token,
        handlers=[
            h := Handler(handler_func, UpdateType.EDITED_MESSAGE),
        ],
    )
    assert dispatcher._routers["edited_message"][-1]._handlers[0] is h


def test_add_handler_channel_post(token, handler_func):
    dispatcher = Dispatcher(
        token,
        handlers=[
            h := Handler(handler_func, UpdateType.CHANNEL_POST),
        ],
    )
    assert dispatcher._routers["channel_post"][-1]._handlers[0] is h


def test_add_handler_edited_channel_post(token, handler_func):
    dispatcher = Dispatcher(
        token,
        handlers=[
            h := Handler(handler_func, UpdateType.EDITED_CHANNEL_POST),
        ],
    )
    assert dispatcher._routers["edited_channel_post"][-1]._handlers[0] is h


def test_add_handler_inline_query(token, handler_func):
    dispatcher = Dispatcher(
        token,
        handlers=[
            h := Handler(handler_func, UpdateType.INLINE_QUERY),
        ],
    )
    assert dispatcher._routers["inline_query"][-1]._handlers[0] is h


def test_add_handler_chosen_inline_result(token, handler_func):
    dispatcher = Dispatcher(
        token,
        handlers=[
            h := Handler(handler_func, UpdateType.CHOSEN_INLINE_RESULT),
        ],
    )
    assert dispatcher._routers["chosen_inline_result"][-1]._handlers[0] is h


def test_add_handler_callback_query(token, handler_func):
    dispatcher = Dispatcher(
        token,
        handlers=[
            h := Handler(handler_func, UpdateType.CALLBACK_QUERY),
        ],
    )
    assert dispatcher._routers["callback_query"][-1]._handlers[0] is h


def test_add_handler_shipping_query(token, handler_func):
    dispatcher = Dispatcher(
        token,
        handlers=[
            h := Handler(handler_func, UpdateType.SHIPPING_QUERY),
        ],
    )
    assert dispatcher._routers["shipping_query"][-1]._handlers[0] is h


def test_add_handler_pre_checkout_query(token, handler_func):
    dispatcher = Dispatcher(
        token,
        handlers=[
            h := Handler(handler_func, UpdateType.PRE_CHECKOUT_QUERY),
        ],
    )
    assert dispatcher._routers["pre_checkout_query"][-1]._handlers[0] is h


def test_add_handler_poll(token, handler_func):
    dispatcher = Dispatcher(
        token,
        handlers=[
            h := Handler(handler_func, UpdateType.POLL),
        ],
    )
    assert dispatcher._routers["poll"][-1]._handlers[0] is h


def test_add_handler_poll_answer(token, handler_func):
    dispatcher = Dispatcher(
        token,
        handlers=[
            h := Handler(handler_func, UpdateType.POLL_ANSWER),
        ],
    )
    assert dispatcher._routers["poll_answer"][-1]._handlers[0] is h


def test_add_handler_my_chat_member(token, handler_func):
    dispatcher = Dispatcher(
        token,
        handlers=[
            h := Handler(handler_func, UpdateType.MY_CHAT_MEMBER),
        ],
    )
    assert dispatcher._routers["my_chat_member"][-1]._handlers[0] is h


def test_add_handler_chat_member(token, handler_func):
    dispatcher = Dispatcher(
        token,
        handlers=[
            h := Handler(handler_func, UpdateType.CHAT_MEMBER),
        ],
    )
    assert dispatcher._routers["chat_member"][-1]._handlers[0] is h


def test_add_handler_chat_join_request(token, handler_func):
    dispatcher = Dispatcher(
        token,
        handlers=[
            h := Handler(handler_func, UpdateType.CHAT_JOIN_REQUEST),
        ],
    )
    assert dispatcher._routers["chat_join_request"][-1]._handlers[0] is h


def test_add_handler_message_reaction(token, handler_func):
    dispatcher = Dispatcher(
        token,
        handlers=[
            h := Handler(handler_func, UpdateType.MESSAGE_REACTION),
        ],
    )
    assert dispatcher._routers["message_reaction"][-1]._handlers[0] is h


def test_add_handler_message_reaction_count(token, handler_func):
    dispatcher = Dispatcher(
        token,
        handlers=[
            h := Handler(handler_func, UpdateType.MESSAGE_REACTION_COUNT),
        ],
    )
    assert dispatcher._routers["message_reaction_count"][-1]._handlers[0] is h


def test_add_handler_chat_boost(token, handler_func):
    dispatcher = Dispatcher(
        token,
        handlers=[
            h := Handler(handler_func, UpdateType.CHAT_BOOST),
        ],
    )
    assert dispatcher._routers["chat_boost"][-1]._handlers[0] is h


def test_add_handler_removed_chat_boost(token, handler_func):
    dispatcher = Dispatcher(
        token,
        handlers=[
            h := Handler(handler_func, UpdateType.REMOVED_CHAT_BOOST),
        ],
    )
    assert dispatcher._routers["removed_chat_boost"][-1]._handlers[0] is h



import unittest.mock as mock

import msgspec
import pytest

from yatbaf import Bot
from yatbaf import OnMessage
from yatbaf import on_message
from yatbaf.exceptions import FrozenInstanceError
from yatbaf.filters import Command


@pytest.mark.asyncio
async def test_polling_update(token, update):
    mark = mock.Mock()

    @on_message(filters=[Command("foo")])
    async def handler(message):
        mark(message)

    bot = Bot(token, handlers=[handler])

    update.message.text = "/foo"
    await bot.process_update(update)

    mark.assert_called_once_with(update.message)


@pytest.mark.asyncio
async def test_webhook_update(token, update):
    mark = mock.Mock()

    @on_message(filters=[Command("foo")])
    async def handler(message):
        mark(message)

    bot = Bot(token, handlers=[handler])

    update.message.text = "/foo"
    webhook_content = msgspec.json.encode(update)
    await bot.process_update(webhook_content)

    mark.assert_called_once_with(update.message)


@pytest.mark.asyncio
async def test_bot_frozen(token):
    router = OnMessage()
    _ = Bot(token, routers=[router])

    with pytest.raises(FrozenInstanceError):

        @router.guard
        async def g(_):
            pass

    with pytest.raises(FrozenInstanceError):

        @router.middleware
        def m(_):
            pass

    with pytest.raises(FrozenInstanceError):

        @router
        async def h(_):
            pass



from __future__ import annotations

from typing import final

from .abc import TelegramType


@final
class MessageId(TelegramType):
    """This object represents a unique message identifier.

    See: https://core.telegram.org/bots/api#messageid
    """

    message_id: int
    """Unique message identifier."""



import pytest

from yatbaf.enums import MarkdownEntity
from yatbaf.helpers import md
from yatbaf.helpers import parse_command_args


@pytest.mark.parametrize(
    "text,result",
    (
        ("/foo", []),
        ("/foo bar", ["bar"]),
        ("/foo@bot bar baz", ["bar", "baz"]),
        ("/foo Bar baZ 123", ["Bar", "baZ", "123"]),
    ),
)
def test_parse_args(text, result):
    assert parse_command_args(text) == result


@pytest.mark.parametrize(
    "text,expect",
    (
        ("foo", "foo"),
        ("foo.", "foo\\."),
        ("*foo*.", "\\*foo\\*\\."),
        ("_*foo__.", "\\_\\*foo\\_\\_\\."),
        ("__foo__", "\\_\\_foo\\_\\_"),
        ("!foo__.", "\\!foo\\_\\_\\."),
        ("foo != bar", "foo \\!\\= bar"),
        ("foo|bar", "foo\\|bar"),
        ("foo {} bar", "foo \\{\\} bar"),
        ("foo > bar", "foo \\> bar"),
        ("#!/usr/foo/bar -baz", "\\#\\!/usr/foo/bar \\-baz"),
        ("#![!_foo(~bar_)`].", "\\#\\!\\[\\!\\_foo\\(\\~bar\\_\\)\\`\\]\\."),
    )
)
def test_markdown_escape(text, expect):
    assert (result := md.escape(text)) == expect, result


@pytest.mark.parametrize(
    "text,expect", (
        ("foo", "foo"),
        ("foo {}", "foo {}"),
        ("foo {0}", "foo {0}"),
    )
)
def test_markdown_escape_fstring(text, expect):
    assert (result := md.escape(text, MarkdownEntity.FSTRING)) == expect, result
    assert result.format(s := "bar") == expect.format(s)


@pytest.mark.parametrize(
    "text,expect",
    (
        ("foo", "foo"),
        ("`foo`", "\\`foo\\`"),
        ("``\\foo\\``", "\\`\\`\\\\foo\\\\\\`\\`"),
    )
)
def test_markdown_escape_code_pre(text, expect):
    assert md.escape(text, MarkdownEntity.PRE) == expect
    assert md.escape(text, MarkdownEntity.CODE) == expect


@pytest.mark.parametrize(
    "text,expect",
    (
        ("foo", "foo"),
        ("http://foo.bar?q=(\\baz)", "http://foo.bar?q=(\\\\baz\\)"),
        ("http://foo.bar?q=\\))\\!baz", "http://foo.bar?q=\\\\\\)\\)\\\\!baz"),
    )
)
def test_markdown_escape_link_emoji(text, expect):
    assert md.escape(text, MarkdownEntity.LINK) == expect
    assert md.escape(text, MarkdownEntity.EMOJI) == expect


def test_markdown_bold():
    assert md.bold("bold") == "*bold*"


def test_markdown_italic():
    assert md.italic("italic") == "_italic_"


def test_markdown_underline():
    assert md.underline("underline") == "__underline__"


def test_markdown_strikethrough():
    assert md.strikethrough("strikethrough") == "~strikethrough~"


def test_markdown_spoiler():
    assert md.spoiler("spoiler") == "||spoiler||"


def test_markdown_url():
    assert (
        md.url("url title", "https://foo.bar") == "[url title](https://foo.bar)"
    )


def test_markdown_mention():
    assert md.mention("user", 1234) == "[user](tg://user?id=1234)"


def test_markdown_emoji():
    assert md.emoji(1234, "emoji") == "![emoji](tg://emoji?id=1234)"


def test_markdown_inline():
    assert md.inline("foo") == "`foo`"


def test_markdown_code():
    assert md.code("print('foo')") == "```\nprint('foo')\n```"
    assert md.code("print('foo')", "python") == "```python\nprint('foo')\n```"



from typing import final

from .abc import TelegramType


@final
class GiveawayCreated(TelegramType):
    """This object represents a service message about the creation of a
    scheduled giveaway. Currently holds no information.

    See: https://core.telegram.org/bots/api#giveawaycreated
    """



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import final

from .abc import TelegramType

if TYPE_CHECKING:
    from yatbaf.enums import ParseMode

    from .link_preview_options import LinkPreviewOptions
    from .message_entity import MessageEntity


@final
class InputTextMessageContent(TelegramType):
    """Represents the content of a text message to be sent as the result of an
    inline query.

    See: https://core.telegram.org/bots/api#inputtextmessagecontent
    """

    message_text: str
    """Text of the message to be sent, 1-4096 characters."""

    parse_mode: ParseMode | None = None
    """*Optional.* Mode for parsing entities in the message text."""

    entities: list[MessageEntity] | None = None
    """*Optional.* List of special entities that appear in message text, which
    can be specified instead of ``parse_mode``.
    """

    link_preview_options: LinkPreviewOptions | None = None
    """*Optional.* Link preview generation options for the message."""



from __future__ import annotations

from typing import final

from .abc import TelegramType


@final
class UsersShared(TelegramType):
    """This object contains information about the users whose identifiers were
    shared with the bot using a :class:`~yatbaf.types.keyboard_buton.KeyboardButtonRequestUsers` button.

    See: https://core.telegram.org/bots/api#usersshared
    """  # noqa: E501

    request_id: int
    """Identifier of the request."""

    user_ids: list[int]
    """Identifiers of the shared users."""



from __future__ import annotations

from typing import final

from .abc import TelegramType
from .user import User


@final
class ProximityAlertTriggered(TelegramType):
    """This object represents the content of a service message, sent whenever
    a user in the chat triggers a proximity alert set by another user.

    See: https://core.telegram.org/bots/api#proximityalerttriggered
    """

    traveler: User
    """User that triggered the alert."""

    watcher: User
    """User that set the alert."""

    distance: int
    """The distance between the users."""



from __future__ import annotations

from typing import final

from .abc import TelegramType
from .chat import Chat
from .chat_boost_source import ChatBoostSource


@final
class ChatBoostRemoved(TelegramType):
    """This object represents a boost removed from a chat.

    See: https://core.telegram.org/bots/api#chatboostremoved
    """

    chat: Chat
    """Chat which was boosted."""

    boost_id: str
    """Unique identifier of the boost."""

    remove_date: int
    """Point in time (Unix timestamp) when the boost was removed."""

    source: ChatBoostSource
    """Source of the removed boost."""



from __future__ import annotations

from typing import final

from yatbaf.enums import Currency

from .abc import TelegramType


@final
class Invoice(TelegramType):
    """This object contains basic information about an invoice.

    See: https://core.telegram.org/bots/api#invoice
    """

    title: str
    """Product name."""

    description: str
    """Product description."""

    start_parameter: str
    """
    Unique bot deep-linking parameter that can be used to generate this invoice.
    """

    currency: Currency
    """Three-letter ISO 4217 currency code."""

    total_amount: int
    """Total price in the smallest units of the currency.

    .. important::

        Integer, not float/double.
    """



import unittest.mock as mock

import pytest

from yatbaf.enums import UpdateType
from yatbaf.exceptions import BotWarning
from yatbaf.filters import Command
from yatbaf.handler import Handler
from yatbaf.router import OnMessage


def test_add_handler(router, handler_func):
    router.add_handler(handler_func)
    assert isinstance(router._handlers[0], Handler)
    assert router._handlers[0]._fn is handler_func
    assert router._handlers[0]._parent is router


def test_add_handler_obj(router, handler_func):
    handler = Handler(handler_func, UpdateType.MESSAGE)
    router.add_handler(handler)
    assert router._handlers[0] is handler


def test_add_handler_type_error(router, handler_func):
    handler = Handler(handler_func, UpdateType.POLL)
    with pytest.raises(ValueError):
        router.add_handler(handler)


def test_add_handler_registered(handler_func):
    handler = Handler(handler_func, UpdateType.MESSAGE)
    _ = OnMessage(handlers=[handler])
    with pytest.raises(ValueError):
        OnMessage(handlers=[handler])


def test_add_handler_registered_same_router(handler_func):
    handler = Handler(handler_func, UpdateType.MESSAGE)
    router = OnMessage(handlers=[handler])
    with pytest.warns(BotWarning):
        router.add_handler(handler)
    assert len(router._handlers) == 1


def test_add_handler_duplicate_func(router, handler_func, filter):
    handler1 = Handler(handler_func, UpdateType.MESSAGE)
    handler2 = Handler(handler_func, UpdateType.MESSAGE, filters=[filter])
    router.add_handler(handler1)
    with pytest.warns(BotWarning):
        router.add_handler(handler2)
    assert len(router._handlers) == 1
    assert router._handlers[0] is handler1


def test_add_handler_duplicate_func1(router, handler_func, filter):
    handler1 = Handler(handler_func, UpdateType.MESSAGE)
    router.add_handler(handler1)
    with pytest.warns(BotWarning):
        router.add_handler(handler_func, filters=[filter])
    assert len(router._handlers) == 1
    assert router._handlers[0] is handler1


def test_add_handler_duplicate_func2(router, handler_func, filter):
    router.add_handler(handler_func)
    with pytest.warns(BotWarning):
        router.add_handler(handler_func, filters=[filter])
    assert len(router._handlers) == 1
    assert not router._handlers[0]._filters


def test_add_handler_decorator(router):

    @router
    async def handler(_):  # noqa: U101
        pass

    assert isinstance(router._handlers[0], Handler)
    assert router._handlers[0]._fn is handler


def test_add_handler_decorator_duplicate(router):

    with pytest.warns(UserWarning):

        @router
        @router
        async def handler(_):  # noqa: U101
            pass

    assert isinstance(router._handlers[0], Handler)
    assert len(router._handlers) == 1
    assert router._handlers[0]._fn is handler


def test_handler_decorator_filter(router):
    filter = Command("foo")

    @router(filters=[filter])
    async def handler(_):  # noqa: U101
        pass

    assert isinstance(router._handlers[0], Handler)
    assert router._handlers[0]._fn is handler
    assert filter in router._handlers[0]._filters


def test_handler_decorator_middleware(router):
    middleware = mock.Mock()

    @router(middleware=[middleware])
    async def handler(_):  # noqa: U101
        pass

    assert isinstance(router._handlers[0], Handler)
    assert router._handlers[0]._fn is handler
    assert middleware in router._handlers[0]._middleware


def test_handler_decorator_middleware_filter_mix(router, filter):
    middleware = mock.Mock()

    @router(
        middleware=[middleware],
        filters=[filter],
    )
    async def handler(_) -> None:  # noqa: U101
        pass

    assert isinstance(router._handlers[0], Handler)
    assert router._handlers[0]._fn is handler
    assert middleware in router._handlers[0]._middleware
    assert filter in router._handlers[0]._filters



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import final

from yatbaf.types import BotCommand

from .abc import TelegramMethod

if TYPE_CHECKING:
    from yatbaf.types import BotCommandScope
    from yatbaf.typing import NoneStr


@final
class GetMyCommands(TelegramMethod[list[BotCommand]]):
    """See :meth:`yatbaf.bot.Bot.get_my_commands`"""

    scope: BotCommandScope | None = None
    language_code: NoneStr = None



from __future__ import annotations

from typing import final

from yatbaf.types import StickerSet

from .abc import TelegramMethod


@final
class GetStickerSet(TelegramMethod[StickerSet]):
    """See :meth:`yatbaf.bot.Bot.get_stickers_set`"""

    name: str



from __future__ import annotations

from typing import final

from .abc import TelegramMethod


@final
class DeleteMessage(TelegramMethod[bool]):
    """See :meth:`yatbaf.bot.Bot.delete_message`"""

    chat_id: str | int
    message_id: int



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import final

from .abc import TelegramMethod

if TYPE_CHECKING:
    from yatbaf.typing import NoneBool


@final
class PromoteChatMember(TelegramMethod[bool]):
    """See :meth:`yatbaf.bot.Bot.promote_chat_member`"""

    chat_id: str | int
    user_id: int
    is_anonymous: NoneBool = None
    can_manage_chat: NoneBool = None
    can_post_messages: NoneBool = None
    can_edit_messages: NoneBool = None
    can_delete_messages: NoneBool = None
    can_manage_video_chats: NoneBool = None
    can_restrict_members: NoneBool = None
    can_promote_members: NoneBool = None
    can_change_info: NoneBool = None
    can_invite_users: NoneBool = None
    can_pin_messages: NoneBool = None
    can_post_stories: NoneBool = None
    can_edit_stories: NoneBool = None
    can_delete_stories: NoneBool = None
    can_manage_topics: NoneBool = None



from __future__ import annotations

from typing import final

from yatbaf.typing import NoneBool
from yatbaf.typing import NoneInt
from yatbaf.typing import NoneStr

from .abc import TelegramType
from .chat import Chat


@final
class Giveaway(TelegramType):
    """This object represents a message about a scheduled giveaway.

    See: https://core.telegram.org/bots/api#giveaway
    """

    chats: list[Chat]
    """The list of chats which the user must join to participate in the
    giveaway.
    """

    winners_selection_date: int
    """Point in time (Unix timestamp) when winners of the giveaway will be
    selected.
    """

    winner_count: int
    """The number of users which are supposed to be selected as winners of the
    giveaway.
    """

    only_new_members: NoneBool = None
    """*Optional.* ``True``, if only users who join the chats after the giveaway
    started should be eligible to win.
    """

    has_public_winners: NoneBool = None
    """*Optional.* ``True``, if the list of giveaway winners will be visible to
    everyone.
    """

    prize_description: NoneStr = None
    """*Optional.* Description of additional giveaway prize."""

    country_codes: list[str] | None = None
    """*Optional.* A list of two-letter ISO 3166-1 alpha-2 country codes
    indicating the countries from which eligible users for the giveaway must
    come. If empty, then all users can participate in the giveaway. Users with
    a phone number that was bought on Fragment can always participate in
    giveaways.
    """

    premium_subscription_month_count: NoneInt = None
    """*Optional.* The number of months the Telegram Premium subscription won
    from the giveaway will be active for.
    """



from __future__ import annotations

from typing import final

from .abc import TelegramMethod


@final
class DeleteMessages(TelegramMethod[bool]):
    """See :meth:`~yatbaf.bot.Bot.delete_messages`"""

    chat_id: int | str
    message_ids: list[int]



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import final

from yatbaf.types import GameHighScore

from .abc import TelegramMethod

if TYPE_CHECKING:
    from yatbaf.typing import NoneInt
    from yatbaf.typing import NoneStr


@final
class GetGameHighScores(TelegramMethod[list[GameHighScore]]):
    """See :meth:`yatbaf.bot.Bot.get_game_high_scores`"""

    user_id: int
    chat_id: NoneInt = None
    message_id: NoneInt = None
    inline_message_id: NoneStr = None



import pytest

from yatbaf.filters import Command


def test_empty():
    with pytest.raises(ValueError):
        Command()


def test_true(message):
    message.text = "/ping"
    assert Command("ping").check(message)


def test_false(message):
    message.text = "/pong"
    assert not Command("ping").check(message)


def test_mix_true(message):
    message.text = "/pong"
    assert Command("ping", "pong").check(message)


def test_text_is_none(message):
    message.text = None
    assert not Command("ping").check(message)


@pytest.mark.parametrize("filter", ["start", "/start", "START", "Start"])
def test_case(message, filter):
    message.text = "/start"
    assert Command(filter).check(message)



from __future__ import annotations

__all__ = (
    "AbstractClient",
    "AbstractRouter",
)

from abc import ABC
from abc import abstractmethod
from typing import TYPE_CHECKING
from typing import Generic

from .typing import UpdateT

if TYPE_CHECKING:
    from collections.abc import AsyncIterator
    from collections.abc import Sequence
    from contextlib import AbstractAsyncContextManager

    from .handler import Handler
    from .middleware import Middleware
    from .models import HTTPResponse
    from .models import Request
    from .models import UpdateInfo
    from .typing import HandlerCallable


class AbstractClient(ABC):
    """Abstract http client."""

    __slots__ = ()

    @abstractmethod
    async def send_post(
        self,
        request: Request,  # noqa: U100
        *,
        timeout: float | None = None  # noqa: U100
    ) -> HTTPResponse[bytes]:
        """Send POST request.

        :param request: :class:`~yatbaf.models.Request` object.
        :param timeout: Request timeout.
        """
        pass

    @abstractmethod
    def file_stream(
        self,
        url: str,  # noqa: U100
        chunk_size: int,  # noqa: U100
    ) -> AbstractAsyncContextManager[HTTPResponse[AsyncIterator[bytes]]]:
        """Download file content.

        :param url: File URL.
        :param chunk_size: Chunk length in bytes.
        """
        pass

    @abstractmethod
    async def close(self) -> None:
        pass


class AbstractRouter(ABC, Generic[UpdateT]):
    __slots__ = ()

    _parent: AbstractRouter[UpdateT] | None
    _middleware: list[Middleware[HandlerCallable[UpdateT]]]

    @property
    @abstractmethod
    def _update_type(self) -> str:
        pass

    @abstractmethod
    def _register_routers(
        self, routers: Sequence[AbstractRouter[UpdateT]]
    ) -> None:
        pass

    @abstractmethod
    def _register_handlers(self, routers: Sequence[Handler[UpdateT]]) -> None:
        pass

    @abstractmethod
    def _on_registration(self) -> None:
        pass

    @abstractmethod
    async def _resolve(self, update: UpdateInfo[UpdateT], /) -> bool:
        pass



from __future__ import annotations

from typing import final

from .abc import TelegramType


@final
class Dice(TelegramType):
    """This object represents an animated emoji that displays a random value.

    See: https://core.telegram.org/bots/api#dice
    """

    emoji: str
    """Emoji on which the dice throw animation is based"""

    value: int
    """Value of the dice, 1-6 for “🎲”, “🎯” and “🎳” base emoji, 1-5 for “🏀”
    and “⚽” base emoji, 1-64 for “🎰” base emoji
    """



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import final

from .abc import TelegramMethod

if TYPE_CHECKING:
    from yatbaf.typing import NoneStr


@final
class SetMyDescription(TelegramMethod[bool]):
    """See :meth:`yatbaf.bot.Bot.set_my_descriptions`"""

    description: NoneStr = None
    language_code: NoneStr = None



from __future__ import annotations

from typing import final

from yatbaf.typing import NoneStr

from .abc import TelegramType


@final
class ForumTopicCreated(TelegramType):
    """This object represents a service message about a new forum topic created
    in the chat.

    See: https://core.telegram.org/bots/api#forumtopiccreated
    """

    name: str
    """Name of the topic."""

    icon_color: int
    """Color of the topic icon in RGB format."""

    icon_custom_emoji_id: NoneStr = None
    """*Optional.* Unique identifier of the custom emoji shown as the topic icon."""  # noqa: E501



from __future__ import annotations

from typing import final

from yatbaf.types import Chat

from .abc import TelegramMethod


@final
class GetChat(TelegramMethod[Chat]):
    """See :meth:`yatbaf.bot.Bot.get_chat`"""

    chat_id: str | int



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import final

from .abc import TelegramMethod

if TYPE_CHECKING:
    from yatbaf.types import ShippingOption
    from yatbaf.typing import NoneStr


@final
class AnswerShippingQuery(TelegramMethod[bool]):
    """See :meth:`yatbaf.bot.Bot.answer_shipping_query`"""

    shipping_query_id: str
    ok: bool
    shipping_options: list[ShippingOption] | None = None
    error_message: NoneStr = None



from __future__ import annotations

from typing import final

from yatbaf.typing import NoneInt
from yatbaf.typing import NoneStr

from .abc import TelegramType
from .photo_size import PhotoSize


@final
class Document(TelegramType):
    """This object represents a general file (as opposed to photos, voice
    messages and audio files).

    See: https://core.telegram.org/bots/api#document
    """

    file_id: str
    """
    Identifier for this file, which can be used to download or reuse the file.
    """

    file_unique_id: str
    """Unique identifier for this file, which is supposed to be the same over
    time and for different bots.

    .. warning::

        Can't be used to download or reuse the file.
    """

    thumbnail: PhotoSize | None = None
    """*Optional.* Document thumbnail as defined by sender."""

    file_name: NoneStr = None
    """*Optional.* Original filename as defined by sender."""

    mime_type: NoneStr = None
    """*Optional.* MIME type of the file as defined by sender."""

    file_size: NoneInt = None
    """*Optional.* File size in bytes."""



from __future__ import annotations

__all__ = [
    "MenuButton",
    "MenuButtonCommands",
    "MenuButtonDefault",
    "MenuButtonWebApp",
]

from typing import ClassVar
from typing import Literal
from typing import TypeAlias
from typing import final

from .abc import TelegramType
from .web_app_info import WebAppInfo


@final
class MenuButtonCommands(TelegramType, tag="commands"):
    """Represents a menu button, which opens the bot's list of commands.

    See: https://core.telegram.org/bots/api#menubuttoncommands
    """

    type: ClassVar[Literal["commands"]] = "commands"
    """Type of the button, must be `commands`"""


@final
class MenuButtonDefault(TelegramType, tag="default"):
    """Describes that no specific value for the menu button was set.

    See: https://core.telegram.org/bots/api#menubuttondefault
    """

    type: ClassVar[Literal["default"]] = "default"
    """Type of the button, must be `default`."""


@final
class MenuButtonWebApp(TelegramType, tag="web_app"):
    """Represents a menu button, which launches a Web App.

    See: https://core.telegram.org/bots/api#menubuttonwebapp
    """

    text: str
    """Text on the button."""

    web_app: WebAppInfo
    """Description of the Web App that will be launched when the user presses
    the button. The Web App will be able to send an arbitrary message on behalf
    of the user using the method :meth:`yatbaf.bot.Bot.answer_web_app_query`.
    """

    type: ClassVar[Literal["web_app"]] = "web_app"
    """Type of the button, must be `web_app`."""


# https://core.telegram.org/bots/api#menubutton
MenuButton: TypeAlias = (
    "MenuButtonCommands | MenuButtonDefault | MenuButtonWebApp"
)



from __future__ import annotations

__all__ = [
    "ChatMember",
    "ChatMemberAdministrator",
    "ChatMemberBanned",
    "ChatMemberLeft",
    "ChatMemberMember",
    "ChatMemberOwner",
    "ChatMemberRestricted",
    "ChatMemberUpdated",
]

# from datetime import datetime
from typing import ClassVar
from typing import Literal
from typing import TypeAlias
from typing import final

from msgspec import field

from yatbaf.typing import NoneBool
from yatbaf.typing import NoneStr

from .abc import TelegramType
from .chat import Chat
from .chat_invite_link import ChatInviteLink
from .user import User


@final
class ChatMemberAdministrator(
    TelegramType, tag="administrator", tag_field="status"
):  # yapf: disable
    """Represents a chat member that has some additional privileges.

    See: https://core.telegram.org/bots/api#chatmemberadministrator
    """

    user: User
    """Information about the user."""

    can_be_edited: bool
    """``True``, if the bot is allowed to edit administrator privileges of that
    user.
    """

    is_anonymous: bool
    """``True``, if the user's presence in the chat is hidden."""

    can_manage_chat: bool
    """``True``, if the administrator can access the chat event log, chat
    statistics, message statistics in channels, see channel members, see
    anonymous administrators in supergroups and ignore slow mode. Implied by
    any other administrator privilege.
    """

    can_delete_messages: bool
    """``True``, if the administrator can delete messages of other users."""

    can_manage_video_chats: bool
    """``True``, if the administrator can manage video chats."""

    can_restrict_members: bool
    """``True``, if the administrator can restrict, ban or unban chat members."""  # noqa: E501

    can_promote_members: bool
    """``True``, if the administrator can add new administrators with a subset
    of their own privileges or demote administrators that they have promoted,
    directly or indirectly (promoted by administrators that were appointed by
    the user).
    """

    can_change_info: bool
    """``True``, if the user is allowed to change the chat title, photo and
    other settings.
    """

    can_invite_users: bool
    """``True``, if the user is allowed to invite new users to the chat."""

    can_post_messages: NoneBool = None
    """*Optional.* ``True``, if the administrator can post in the channel.

    .. note::

        Channels only.
    """

    can_edit_messages: NoneBool = None
    """*Optional.* ``True``, if the administrator can edit messages of other
    users and can pin messages.

    .. note::

        Channels only.
    """

    can_pin_messages: NoneBool = None
    """*Optional.* ``True``, if the user is allowed to pin messages.

    .. note::

        Groups and Supergroups only.
    """

    can_manage_topics: NoneBool = None
    """*Optional.* ``True``, if the user is allowed to create, rename, close,
    and reopen forum topics.

    .. note::

        Supergroups only.
    """

    can_post_stories: NoneBool = None
    """*Optional.* ``True``, if the administrator can post stories in the
    channel.

    .. note::

        Channels only.
    """

    can_edit_stories: NoneBool = None
    """Optional. ``True``, if the administrator can edit stories posted by other
    users.

    .. note::

        Channels only.
    """

    can_delete_stories: NoneBool = None
    """*Optional.* ``True``, if the administrator can delete stories posted by
    other users.

    .. note::

        Channels only.
    """

    custom_title: NoneStr = None
    """*Optional.* Custom title for this user."""

    status: ClassVar[Literal["administrator"]] = "administrator"
    """The member's status in the chat, always `administrator`."""


@final
class ChatMemberBanned(TelegramType, tag="kicked", tag_field="status"):
    """Represents a chat member that was banned in the chat and can't return to
    the chat or view chat messages.

    See: https://core.telegram.org/bots/api#chatmemberbanned
    """

    user: User
    """Information about the user."""

    until_date: int  # datetime
    """Date when restrictions will be lifted for this user; unix time. If 0,
    then the user is banned forever.
    """

    status: ClassVar[Literal["kicked"]] = "kicked"
    """The member's status in the chat, always `kicked`."""


@final
class ChatMemberLeft(TelegramType, tag="left", tag_field="status"):
    """Represents a chat member that isn't currently a member of the chat, but
    may join it themselves.

    See: https://core.telegram.org/bots/api#chatmemberleft
    """

    user: User
    """Information about the user."""

    status: ClassVar[Literal["left"]] = "left"
    """The member's status in the chat, always `left`."""


@final
class ChatMemberMember(TelegramType, tag="member", tag_field="status"):
    """Represents a chat member that has no additional privileges or
    restrictions.

    See: https://core.telegram.org/bots/api#chatmembermember
    """

    user: User
    """Information about the user."""

    status: ClassVar[Literal["member"]] = "member"
    """The member's status in the chat, always `member`."""


@final
class ChatMemberOwner(TelegramType, tag="creator", tag_field="status"):
    """Represents a chat member that owns the chat and has all administrator
    privileges.

    See:  https://core.telegram.org/bots/api#chatmemberowner
    """

    user: User
    """Information about the user."""

    is_anonymous: bool
    """``True``, if the user's presence in the chat is hidden."""

    custom_title: NoneStr = None
    """*Optional.* Custom title for this user."""

    status: ClassVar[Literal["creator"]] = "creator"
    """The member's status in the chat, always `creator`."""


@final
class ChatMemberRestricted(TelegramType, tag="restricted", tag_field="status"):
    """Represents a chat member that is under certain restrictions in the chat.
    Supergroups only.

    See: https://core.telegram.org/bots/api#chatmemberrestricted
    """

    user: User
    """Information about the user."""

    is_member: bool
    """``True``, if the user is a member of the chat at the moment of the
    request.
    """

    can_send_messages: bool
    """``True``, if the user is allowed to send text messages, contacts,
    invoices, locations and venues.
    """

    can_send_audios: bool
    """``True``, if the user is allowed to send audios."""

    can_send_documents: bool
    """``True``, if the user is allowed to send documents."""

    can_send_photos: bool
    """``True``, if the user is allowed to send photos."""

    can_send_videos: bool
    """``True``, if the user is allowed to send videos."""

    can_send_video_notes: bool
    """``True``, if the user is allowed to send video notes."""

    can_send_voice_notes: bool
    """``True``, if the user is allowed to send voice notes."""

    can_send_polls: bool
    """``True``, if the user is allowed to send polls."""

    can_send_other_messages: bool
    """``True``, if the user is allowed to send animations, games, stickers
    and use inline bots.
    """

    can_add_web_page_previews: bool
    """``True``, if the user is allowed to add web page previews to their
    messages.
    """

    can_change_info: bool
    """``True``, if the user is allowed to change the chat title, photo and
    other settings.
    """

    can_invite_users: bool
    """``True``, if the user is allowed to invite new users to the chat."""

    can_pin_messages: bool
    """``True``, if the user is allowed to pin messages."""

    can_manage_topics: bool
    """``True``, if the user is allowed to create forum topics."""

    until_date: int  # TODO: datetime
    """Date when restrictions will be lifted for this user; unix time. If 0,
    then the user is restricted forever.
    """

    status: ClassVar[Literal["restricted"]] = "restricted"
    """The member's status in the chat, always `restricted`."""


ChatMember: TypeAlias = (
    ChatMemberAdministrator
    | ChatMemberBanned
    | ChatMemberLeft
    | ChatMemberMember
    | ChatMemberOwner
    | ChatMemberRestricted
)


@final
class ChatMemberUpdated(TelegramType):
    """This object represents changes in the status of a chat member.

    See: https://core.telegram.org/bots/api#chatmemberupdated
    """

    chat: Chat
    """Chat the user belongs to."""

    from_: User = field(name="from")
    """Performer of the action, which resulted in the change."""

    date: int  # TODO: datetime
    """Date the change was done in Unix time."""

    old_chat_member: ChatMember
    """Previous information about the chat member."""

    new_chat_member: ChatMember
    """New information about the chat member."""

    invite_link: ChatInviteLink | None = None
    """*Optional.* Chat invite link, which was used by the user to join the
    chat; for joining by invite link events only.
    """

    via_chat_folder_invite_link: NoneBool = None
    """*Optional.* ``True``, if the user joined the chat via a chat folder
    invite link.
    """



from __future__ import annotations

from typing import final

from .abc import TelegramType


@final
class BotCommand(TelegramType):
    """This object represents a bot command.

    See: https://core.telegram.org/bots/api#botcommand
    """

    command: str
    """Text of the command.

    .. note::

        Can contain only lowercase English letters, digits and underscores.
    """

    description: str
    """Description of the command."""
